"""
Train an autoencoder to reduce dimensionality of input features.
"""
import logging
import math
import os
import pickle

import numpy as np
import skfuzzy as fuzz
from sklearn.preprocessing import MinMaxScaler
from sklearn.impute import SimpleImputer
from sklearn.pipeline import make_pipeline
from sklearn.model_selection import train_test_split
from sklearn.base import TransformerMixin, BaseEstimator
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset
from typing import List
import torch.nn.functional as F
import xarray as xr
import yaml

from zizou import AnomalyDetectionBaseClass, get_data
import zizou.zizou_logging

logger = logging.getLogger(__name__)


class TiedAutoEncoder(nn.Module):
    """
    Applies an tied-weight autoencoder to the incoming data.
    From https://gist.github.com/northanapon/375e17fb395391c144deff20914e51df
    Parameters
    -------------
    in_features : int
        size of each input sample.
    h_features : List[int]
        a list of size of each layer for the encoder and
        the reverse for the decoder.
    activation : function, optional
        an activation function to apply for each layer 
        (the default is torch.tanh).
    """
    
    def __init__(
            self, in_features: int, h_features: List[int], 
            activation=F.relu):
        """Create an autoencoder."""
        super().__init__()
        self.in_features = in_features
        self.h_features = h_features
        self.encoded_features = h_features[-1]
        self.activation = activation
        self.weights = nn.ParameterList([])
        self.enc_biases = nn.ParameterList([])
        self.dec_biases = nn.ParameterList([])
        in_dim = in_features
        for h in h_features:
            self.weights.append(nn.Parameter(torch.DoubleTensor(h, in_dim)))
            self.enc_biases.append(nn.Parameter(torch.DoubleTensor(h)))
            in_dim = h
        for h in reversed(h_features[:-1]):
            self.dec_biases.append(nn.Parameter(torch.DoubleTensor(h)))
        self.dec_biases.append(nn.Parameter(torch.DoubleTensor(in_features)))
        self.reset_parameters()
        self.loss = torch.nn.MSELoss(reduction='none')
            
    def forward(self, x: torch.DoubleTensor):
        """Return result of encoding and decoding."""
        dec = self.encode(x)
        return self.decode(dec)
    
    def encode(self, x: torch.DoubleTensor):
        """Return encoded data."""
        o = x
        for w, b in zip(self.weights[:-1], self.enc_biases[:-1]):
            o = F.linear(o, w, b)
            o = self.activation(o)
        return F.linear(o, self.weights[-1], self.enc_biases[-1])
    
    def decode(self, o: torch.DoubleTensor):
        """Return decoded data."""
        r_weights = list(reversed(self.weights))
        for w, b in zip(r_weights[:-1], self.dec_biases[:-1]):
            o = F.linear(o, w.t(), b)
            o = self.activation(o)
        return F.linear(o, r_weights[-1].t(), self.dec_biases[-1])
    
    def reset_parameters(self):
        """Reset linear module parameters (from nn.Linear)."""
        for w, b in zip(self.weights, self.enc_biases):
            nn.init.kaiming_uniform_(w, a=math.sqrt(5))
            if b is not None:
                fan_in, _ = nn.init._calculate_fan_in_and_fan_out(w)
                bound = 1 / math.sqrt(fan_in)
                nn.init.uniform_(b, -bound, bound)
        for w, b in zip(self.weights, self.dec_biases):
            if b is not None:
                fan_in, _ = nn.init._calculate_fan_in_and_fan_out(w.t())
                bound = 1 / math.sqrt(fan_in)
                nn.init.uniform_(b, -bound, bound)


class FuzzyCluster(TransformerMixin, BaseEstimator):
    def __init__(self, n_clusters=5, error=0.005, maxiter=1000,
                 init=None):
        self.n_clusters_ = n_clusters
        self.error_ = error
        self.maxiter_ = maxiter
        self.init_ = init
        self.clusterscaler = MinMaxScaler()
        self._is_fitted = False

    def fit(self, data, scalerfile, clustercenterfile):
        try:
            with open(scalerfile, 'rb') as fh:
                self.clusterscaler = pickle.load(fh)
            self.cntr_ = np.load(clustercenterfile)
            self._is_fitted = True
            logging.info("Loaded cluster model.")
        except FileNotFoundError:
            data_scaled = self.clusterscaler.fit_transform(data)
            self.cntr_, u, u0, d, jm, p, fpc = fuzz.cluster.cmeans(
            data_scaled.T, self.n_clusters_, 2, error=0.005, maxiter=1000, init=None)
            with open(scalerfile, 'wb') as fh:
                pickle.dump(self.clusterscaler, fh)
            np.save(clustercenterfile, self.cntr_) 
            self._is_fitted = True

    def transform(self, data):
        assert self._is_fitted
        data_scaled = self.clusterscaler.transform(data)
        u, u0, d, jm, p, fpc = fuzz.cluster.cmeans_predict(
        data_scaled.T, self.cntr_, 2, error=0.005, maxiter=1000)
        return u

    def fit_transform(self, data, scalerfile, clustercenterfile):
        self.fit(data, scalerfile, clustercenterfile)
        return self.transform(data)


class VUMTData(Dataset):
    def __init__(self, data, transform=None):
        self.data = data
        self.transform = transform

    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, idx):
        sample = self.data[idx, :]
        target = self.data[idx, :]
        if self.transform is not None:
            sample = self.transform(sample)
            target = self.transform(target)
        return sample, target


class AutoEncoder(AnomalyDetectionBaseClass):
    features = ['rsam',
                'dsar',
                'central_freq',
                'predom_freq',
                'bandwidth',
                'rsam_energy_prop',
                'sonogram']

    transform_dict = {'rsam': 'log'}

    stack_dict = {'dsar': '2D',
                  'central_freq': '1H',
                  'predom_freq': '1H',
                  'variance': '1H',
                  'bandwidth': '1H'}

    files = dict(featurefile="autoencoder_features.nc",
                 modelfile="autoencoder_weights.pth",
                 scalermodelfile="autoencoder_scaler.pkl",
                 scalerfile="cluster_scaler.pkl",
                 clustercenterfile="cluster_center.npy")
  
    def __init__(self, configfile):
        try:
            with open(configfile, 'r') as fh:
                c = yaml.safe_load(fh)
        except OSError:
            c = yaml.safe_load(configfile)
        self.layers_ = c['autoencoder'].get('layers', [2000, 500, 200, 6])
        self.epochs_ = c['autoencoder'].get('epochs', 5)                       
        self.patience_ = c['autoencoder'].get('patience', 10)
        self.device_ = c['autoencoder'].get('device', 'cuda') 
 
        super(AutoEncoder, self).__init__(self.features, self.stack_dict,
                                          self.transform_dict)
        self.device_ = (self.device_
                        if torch.cuda.is_available()
                        else 'mps'
                        if torch.backends.mps.is_available()
                        else 'cpu')
        logger.info(f"Using {self.device_} device") 
        self._is_fitted = False
        imputer = SimpleImputer(missing_values=np.nan, strategy='median')
        scaler = MinMaxScaler()
        self.transformer = make_pipeline(imputer, scaler)
        self.cluster_ = FuzzyCluster()
        self.classifications = None
   
    def checkpoint(self, model, filename):
        torch.save(model.state_dict(), filename)
    
    def resume(self, model, filename):
        model.load_state_dict(torch.load(filename))

    def train(self, dataloader, optimizer):
        size = len(dataloader.dataset)
        self.model_.train()
        for batch, (X, y) in enumerate(dataloader):
            X, y = X.to(self.device_), y.to(self.device_)

            # Compute prediction error
            pred = self.model_(X)
            loss = self.model_.loss(pred, y).sum(axis=1).mean()

            # Backpropagation
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()

            if batch % 100 == 0:
                loss, current = loss.item(), (batch + 1) * len(X)
                logger.info(f"Training loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")
    
    def test(self, dataloader, loss_fn):
        num_batches = len(dataloader)
        self.model_.eval()
        test_loss = 0
        with torch.no_grad():
            for X, y in dataloader:
                X, y = X.to(self.device_), y.to(self.device_)
                pred = self.model_(X)
                test_loss += loss_fn(pred, y).sum(axis=1).mean()
        test_loss /= num_batches
        self.loss_ = test_loss
        logger.info(f"Test loss: {test_loss:>8f} \n")
        return test_loss
 
    def fit(self, fq):
        autoencoder_ = TiedAutoEncoder(14, self.layers_)
        self.model_ = autoencoder_.to(self.device_)
        logger.info(self.model_)
        try:
            self.model_.load_state_dict(torch.load(os.path.join(fq.sitedir, self.files['modelfile'])))
            with open(os.path.join(fq.sitedir, self.files['scalermodelfile']), 'rb') as fh:
                self.transformer = pickle.load(fh)
            self._is_fitted = True
            logging.info("Loaded pretrained model.")
        except FileNotFoundError:
            optimizer = torch.optim.SGD(self.model_.parameters(), lr=1e-3)
            data = self.get_features(fq, os.path.join(fq.sitedir, self.files['featurefile'])).values
            logger.info(f"Scaling features for dataset with shape {data.shape}")
            # Scale values and fill gaps.
            data_norm = self.transformer.fit_transform(data)
            with open(os.path.join(fq.sitedir, self.files['scalermodelfile']), 'wb') as fh:
                pickle.dump(self.transformer, fh)
            logger.info("Splitting dataset...")
            data_norm_train, data_norm_test = train_test_split(data_norm,
                                                            test_size=.2,
                                                            train_size=.8)
            train_data = VUMTData(data_norm_train, transform=torch.from_numpy)
            test_data = VUMTData(data_norm_test, transform=torch.from_numpy)
            train_dataloader = DataLoader(train_data, batch_size=64, shuffle=True)
            test_dataloader = DataLoader(test_data, batch_size=64, shuffle=True)
            logger.info("Starting training...")
            best_loss = 1e30
            best_epoch = 0
            for i in range(self.epochs_):
                logger.info(f"Epoch {i+1}")
                self.train(train_dataloader, optimizer)
                loss = self.test(test_dataloader, self.model_.loss)
                if loss < best_loss:
                    best_loss = loss
                    best_epoch = i
                    self.checkpoint(self.model_, os.path.join(fq.sitedir, self.files['modelfile'])) 
                elif i - best_epoch > self.patience_:
                    logger.info(f"Early stopped training at epoch {i+1}")
                    break
            self._is_fitted = True

    def transform(self, fq, cluster=True):
        assert self._is_fitted
        feats = self.get_features(fq)
        dates = feats['datetime']
        data = feats.values
        data_norm = self.transformer.transform(data)
        transform_data = VUMTData(data_norm,
                                  transform=torch.from_numpy)
        with torch.no_grad():
            X = transform_data[:][0]
            X = X.to(self.device_)
            encoded_data = self.model_.encode(X)
        encoded_data = encoded_data.cpu().numpy()
        if cluster:
            clustered_data = self.cluster_.fit_transform(encoded_data,
                                                         os.path.join(fq.sitedir, self.files['scalerfile']),
                                                         os.path.join(fq.sitedir, self.files['clustercenterfile']))
            cluster_names = list(range(self.cluster_.n_clusters_))
            xda = xr.DataArray(clustered_data,
                                coords=[cluster_names, dates],
                                dims=['cluster', 'datetime'])
            return xr.Dataset({'autoencoder': xda})
        xda = xr.DataArray(encoded_data.T,
                           coords=[list(range(encoded_data.shape[1])),
                                  dates],
                           dims=['autodim', 'datetime'])
        return xr.Dataset({'autoencoder': xda})

    def fit_transform(self, fq, cluster=True):
        self.fit(fq)
        self.classifications = self.transform(fq, cluster=cluster)
        return self.classifications



"""
Functionality common to several modules.
"""
from datetime import datetime
import logging
import math as M
import os

import h5netcdf
from cftime import num2date, date2num, date2index
import numpy as np
from obspy import UTCDateTime, Trace
import pandas as pd
from scipy.signal import chirp, sweep_poly
import xarray as xr
from warnings import filterwarnings


logger = logging.getLogger(__name__)


def stride_windows(x, n, noverlap):
    """
    Get all windows of x with length n as a single array,
    using strides to avoid data duplication.

    :param x: 1D array or sequence containing the data.
    :type x: :class:`~numpy.ndarray`
    :param n: The number of data points in each window.
    :type n: int
    :param noverlap: The overlap between adjacent windows.
                     Default is 0 (no overlap)
    :type noverlap: int
    :return: Array with the windowed data in columns.
    :rtype: :class:`~numpy.ndarray`
    """

    if noverlap >= n:
        raise ValueError('noverlap must be less than n')
    if n < 1:
        raise ValueError('n cannot be less than 1')

    if n == 1 and noverlap == 0:
        return x[np.newaxis]
    if n > x.size:
        raise ValueError('n cannot be greater than the length of x')

    # np.lib.stride_tricks.as_strided easily leads to memory corruption for
    # non integer shape and strides, i.e. noverlap or n. See #3845.
    noverlap = int(noverlap)
    n = int(n)

    step = n - noverlap
    shape = (n, (x.shape[-1] - noverlap) // step)
    strides = (x.strides[0], step * x.strides[0])
    return np.lib.stride_tricks.as_strided(x, shape=shape, strides=strides,
                                           writeable=False)


def demean(x, axis=None):
    """
    Return x minus the mean(x).

    :param x: array or sequence containing the data can
              have any dimensionality
    :type x: :class:`~numpy.ndarray`
    :param axis: The axis along which to take the mean. See numpy.mean
                 for a description of this argument.
    :type axis: int
    :return: input minus the arithmetic mean along the specified axis.
    :rtype: :class:`~numpy.ndarray`
    """
    x = np.asarray(x)

    if axis is not None and axis + 1 > x.ndim:
        raise ValueError('axis(=%s) out of bounds' % axis)

    return x - np.nanmean(x, axis=axis, keepdims=True)


def apply_hanning(x, return_window=None):
    """
    Apply a hanning window to the given 1D or 2D array along
    the given axis.

    :param x: 1-D or 2-D array or sequence containing the data
    :type x: :class:`numpy.ndarray`
    :param return_window: If true, also return the 1D values of the window
                          that was applied.
    :type return_window: bool
    :return: **wtr[, wvals]** input multiplied with the hanning
             window function, hanning window function
    """
    x = np.asarray(x)

    if x.ndim < 1 or x.ndim > 2:
        raise ValueError('only 1D or 2D arrays can be used')

    xshapetarg = x.shape[0]
    windowVals = np.hanning(xshapetarg) * np.ones(xshapetarg, dtype=x.dtype)

    if x.ndim == 1:
        if return_window:
            return windowVals * x, windowVals
        else:
            return windowVals * x

    windowVals = windowVals[:, np.newaxis]

    if return_window:
        return windowVals * x, windowVals
    else:
        return windowVals * x


def window_array(x, nwin, noverlap,
                 remove_mean=True, taper=True,
                 padval=0, return_window=True):
    """
    Take a 1-D numpy array and devide it into (overlapping)
    windows.

    :param x: 1D array containing the data.
    :type x: :class:`~numpy.ndarray`
    :param n: The number of data points in each window. If this
              is not a power of 2, the length of each window will be
              the nearest power of 2.
    :type n: int
    :param noverlap: The number of overlapping points between
                     adjacent windows.
    :type noverlap: int
    :param remove_mean: Remove the individual mean from each
                        window.
    :type remove_mean: bool
    :param taper: Apply a hanning taper to each window.
    :type taper: bool
    :param return_window: If taper is 'True', return the values
                          of the hanning taper.
    :type return_window: bool
    :param padval: If this is not 'None', padval will be used to
                   pad the input array such that its length is
                   a multiple of the window length.
    :type padval: float, class:`numpy.nan`, or None
    :return: Array with the windowed data in columns.
    :rtype: :class:`~numpy.ndarray`
    """
    x = x.astype('float')
    npts = x.size
    if nwin > npts:
        msg = "window length can't be greater than array length"
        raise ValueError(msg)

    # step length that the window is advanced by
    step = nwin - noverlap
    # compute the remainder of the trace after dividing it
    # into windows
    rest = (npts - nwin) % step

    # if the remainder is >0 pad with zeros
    if rest > 0:
        npts_pad = npts + (step - rest)
        x = np.resize(x, int(npts_pad))
        x[npts:] = padval

    result = stride_windows(x, nwin, noverlap)
    if remove_mean:
        result = demean(result, axis=0)
    if taper:
        result, windowVals = apply_hanning(result, return_window=True)
        if return_window:
            return result, windowVals
    return result


def trace_window_times(trace, window_len, min_len=0.8):
    """Subdivide an `obspy.core.trace.Trace` into windows of `window_len`
    length and iterate over window start & end times.

    :param trace: The data to be windowed.
    :type trace: :class:`obspy.core.trace.Trace`
    :param window_len: The window length in seconds.
    :type window_len: float
    :param min_len: Windows with length < `min_len * window_len` are
        skipped, defaults to 0.8
    :type min_len: float, optional
    :yield: The window start & end times
    :rtype: :class:`obspy.core.utcdatetime.UTCDateTime`,
        :class:`obspy.core.utcdatetime.UTCDateTime`,
    """

    min_len = float(min_len)
    if not 0 < min_len <= 1.0:
        raise ValueError("min_len <= 0 or > 1.")
    min_len = min_len * window_len

    if (trace.stats.endtime - trace.stats.starttime) < min_len:
        raise ValueError("window length > total trace data.")

    t0 = trace.stats.starttime
    while t0 < trace.stats.endtime:
        t1 = min(t0 + window_len, trace.stats.endtime)
        this_window_len = t1 - t0
        if this_window_len >= min_len:
            if this_window_len < window_len:
                t0 = t1 - window_len
            yield t0, t1

        t0 += window_len


def trace_window_data(trace, window_len, min_len=0.8):
    """Subdivide an `obspy.core.trace.Trace` into windows of `interval`
    length and iterate over windows.

    :param trace: The data to be windowed.
    :type trace: :class:`obspy.core.trace.Trace`
    :param window_len: The window length in seconds.
    :type window_len: float
    :param min_len: Windows with length < `min_length * window_len` are
        skipped, defaults to 0.8
    :type min_len: float, optional
    :yield: A trace data windowNone
    :rtype: `obspy.core.trace.Trace`
    """

    for t0, t1 in trace_window_times(trace, window_len, min_len):
        yield trace.slice(t0, t1, nearest_sample=False)


def apply_freq_filter(
    data, filtertype, filterfreq, corners=4, zerophase=False, **options
):
    """Apply frequency filter to seismic data.

    Filter parameters are taken from instance attributes
    `filtertype` and `filterfreq`.

    :param data: The seismic data to be filtered.
    :type data: :class:`obspy.core.trace.Trace` or
        :class:``obspy.core.stream.Stream`
    :param corners: Filter corners / order, defaults to 4.
    :type corners: int
    :param zerophase: If True, apply filter once forwards and once
        backwards. Defaults to False.
    :type zerophase: bool
    :param options: Passed to underlying obspy filter methods.
    """
    f_min, f_max = filterfreq
    options.update({"zerophase": zerophase, "corners": corners})

    if filtertype in ["lp", "lowpass"]:
        if f_max is None:
            raise ValueError(
                "Define upper frequency cutoff for low-pass filter."
            )
        data.filter("lowpass", freq=f_max, **options)

    elif filtertype in ["hp", "highpass"]:
        if f_min is None:
            raise ValueError("Define lower frequency cutoff for high-pass filter.")
        data.filter("highpass", freq=f_min, **options)

    elif filtertype in ["bp", "bandpass"]:
        if f_min is None or f_max is None:
            raise ValueError(
                "Define upper and lower requency cutoff for band-pass filter."
            )
        data.filter("bandpass", freqmin=f_min, freqmax=f_max, **options)

    elif filtertype is not None:
        raise ValueError(f"Unrecognised filtertype '{filtertype}'")


def round_time(time, interval):
    """
    Find closest multiple of interval to time.

    :param time: Time to round
    :type time: :class:`obspy.UTCDateTime`
    :param interval: Interval length in seconds
    :type interval: int
    :returns: The new time that is a multiple of interval
    :rtype: :class:`obspy.UtCDateTime`
    """
    tmp = (time + 0.5 * interval).timestamp % interval
    ntime = (time + 0.5 * interval).timestamp - tmp
    return UTCDateTime(ntime)


def xarray2hdf5(xArray, fdir, rootGroupName="original", timedim="datetime"):
    """
    Store an xarray dataset as an HDF5 file.

    :param xArray: Data to store. 
    :type xArray: :class:`xarray.Dataset`
    :param fdir: Directory to store data under.
    :type fdir: str
    :param rootGroupName: Hdf5 group name.
    :type rootGroupName: str
    :param timedim: Name of time dimension.
    :type timedim: str
    """
    filterwarnings(action='ignore', category=DeprecationWarning,
               message='`np.bool` is a deprecated alias')

    for featureName in list(xArray.data_vars.keys()):
        h5file = os.path.join(fdir, featureName +'.nc')

        mode = 'a' if os.path.isfile(h5file) else 'w'
        
        with h5netcdf.File(h5file, mode) as h5f:
            try:
                rootGrp = _create_h5_Structure(rootGroupName, featureName, h5f, xArray)
            except ValueError: # group already exists, append
                rootGrp = h5f[rootGroupName]

            # determine indices
            new_time = date2num(xArray[timedim].values.astype('datetime64[us]').astype(datetime),
                                units=rootGrp[timedim].attrs['units'],
                                calendar=rootGrp[timedim].attrs['calendar'])
            dt = (np.diff(xArray['datetime'])/np.timedelta64(1, 'h'))[0]
            t0 = date2num(np.datetime64(rootGrp.attrs['starttime']).astype('datetime64[us]').astype(datetime),
                          units=rootGrp[timedim].attrs['units'],
                          calendar=rootGrp[timedim].attrs['calendar'])
            indices = np.rint((new_time - t0)/dt).astype(int)
            assert np.all(indices >= 0)
            times = rootGrp[timedim]
            newsize = indices[-1] + 1
            if newsize > times.shape[0]:
                rootGrp.resize_dimension(timedim, newsize)
            times[:] = t0 + np.arange(times.shape[0])*dt
            data = rootGrp[featureName]
            if len(data.shape) > 1:
                data[:, indices] = xArray[featureName].values
            else:
                data[indices] = xArray[featureName].values
            rootGrp.attrs['endtime'] = str(num2date(times[-1], units=rootGrp[timedim].attrs['units'],
                                                    calendar=rootGrp[timedim].attrs['calendar']))
            try:
                _setMetaInfo(featureName, h5f, xArray)
            except KeyError as e:
                logging.warning(f"Could not set all meta info for {featureName}: {e}")


def _create_h5_Structure(defaultGroupName, featureName, h5f, xArray):
    rootGrp = h5f.create_group(defaultGroupName)
    for label, size in xArray.dims.items(): 
        _setAttributes(label, size, rootGrp, xArray)
    # Note: xArray.dims returns a dictionary of dimensions that are not necesarily
    # in the right order; xArray[featureName].dims returns a tuple with dimension
    # names in the correct order
    rootGrp.create_variable(featureName, tuple(xArray[featureName].dims), dtype=float, fillvalue=0.)
    return rootGrp


def _setAttributes(label, size, rootGrp, xArray):
    if np.issubdtype(xArray[label].dtype, np.datetime64):
        starttime = str(xArray[label].values[0].astype('datetime64[us]').astype(datetime))
        rootGrp.dimensions[label] = None
        coordinates = rootGrp.create_variable(label, (label,), float)
        coordinates.attrs['units'] = 'hours since 1970-01-01 00:00:00.0'
        coordinates.attrs['calendar'] = 'gregorian'
        rootGrp.attrs['starttime'] = starttime
    else:
        rootGrp.dimensions[label] = size 
        coordinates = rootGrp.create_variable(label, (label,), float)
        coordinates[:] = xArray[label].values
        

def _setMetaInfo(featureName, h5f, xArray):
    h5f.attrs['station'] = xArray.attrs['station']
    h5f.attrs['latitude'] = -42
    h5f.attrs['longitude'] = 168
    h5f.attrs['datatype'] = featureName
    

def generate_test_data(dim=1, ndays=30, nfreqs=10,
                       tstart=datetime.utcnow(),
                       feature_name=None,
                       freq_name=None):
    """
    Generate a 1D or 2D feature for testing.
    """
    assert dim < 3
    assert dim > 0

    nints = ndays * 6 * 24
    dates = pd.date_range(tstart.strftime('%Y-%m-%d'), freq='10min', periods=nints)
    rs = np.random.default_rng(42)
    # Random walk as test signal
    data = np.abs(np.cumsum(rs.normal(0, 8., len(dates))))
    if dim == 2:
        data = np.tile(data, (nfreqs, 1))
    # Add 10% NaNs
    idx_nan = rs.integers(0, nints-1, int(0.1*nints))
    if dim == 1:
        data[idx_nan] = np.nan
        if feature_name is None:
            feature_name = 'rsam'
        xrd = xr.Dataset({feature_name: xr.DataArray(data, coords=[dates], dims=['datetime'])})
    if dim == 2:
        data[:, idx_nan] = np.nan
        freqs = np.arange(nfreqs)
        if feature_name is None:
            feature_name = 'ssam'
        if freq_name is None:
            freq_name = 'frequency'
        xrd = xr.Dataset({feature_name: xr.DataArray(data, coords=[freqs, dates], dims=[freq_name, 'datetime'])})
    xrd.attrs['starttime'] = UTCDateTime(dates[0]).isoformat()
    xrd.attrs['endtime'] = UTCDateTime(dates[-1]).isoformat()
    xrd.attrs['station'] = 'MDR'
    return xrd


def test_signal(nsec=3600, sampling_rate=100.,
                frequencies=[0.1, 3.0, 10.0],
                amplitudes=[0.1, 1.0, 0.7],
                phases=[0.0, np.pi*0.25, np.pi],
                offsets=[0., 0., 0.],
                starttime=UTCDateTime(1970, 1, 1),
                gaps=False,  noise=True, noise_std=.5,
                sinusoid=True,addchirp=True,
                network='NZ', station='BLUB', location='',
                channel='HHZ'):
    """
    Produce a test signal for which we know where the peaks
    are in the spectrogram.

    :param nsec: Length of the trace in seconds.
    :type nsec: int
    :param sampling_rate: Sampling rate of the signal in Hz.
    :type sampling_rate: float
    :param starttime: Starttime of the trace.
    :type starttime: :class:`~obspy.UTCDateTime`
    :param gaps: If 'True' add gaps to the test signal.
    :type gaps: boolean
    :param noise_std: Standard deviation of the noise.
    :type noise_std: float
    :param sinusoid: Add a signal with a sinusoidal frequency change.
    :type sinusoid: bool
    :param station: Station name of the returned trace.
    :type station: str
    :param location: Location of the returned trace.
    :type location: str
    :return: Time series of test data
    :rtype: :class:`~obspy.Trace`
    """
    t = np.arange(0, nsec, 1 / sampling_rate)
    signals = []
    # Some constant signals with different phases
    for f, A, ph, dt in zip(frequencies, amplitudes, phases, offsets):
        _s = np.zeros(t.size)
        dt_idx = int(dt*sampling_rate)
        _s[dt_idx:]= A * np.sin(2.0 * np.pi * f * t[dt_idx:] + ph)
        signals.append(_s)
    if addchirp:
        # Frequency-swept signals
        # Chirp
        s4 = 0.8 * chirp(t, 5, t[-1], 15, method='quadratic')
        signals.append(s4)

    if sinusoid:
        vals = [6]
        for k in range(0, 7):
            vals.append(2*(-1)**k*np.power(2*np.pi/nsec, 2*k+1)/M.factorial(2*k+1))
            vals.append(0)
        p = np.poly1d(np.array(vals[::-1]))
        s5 = sweep_poly(t, p)
        signals.append(s5)

    # add some noise
    if noise:
        rs = np.random.default_rng(42)
        noise = rs.normal(loc=0., scale=noise_std, size=t.size)
        for s in signals:
            noise += s
        signal = noise
    else:
        signal = signals[0]
        for s in signals[1:]:
            signal += s
    stats = {'network': network, 'station': station, 'location': location,
             'channel': channel, 'npts': len(signal),
             'sampling_rate': sampling_rate,
             'mseed': {'dataquality': 'D'}}
    stats['starttime'] = starttime
    stats['endtime'] = stats['starttime'] + nsec
    if gaps:
        p1 = (0.27, 0.33)
        p2 = (0.69, 0.83)
        p3 = (.95, 1)
        for pmin, pmax in [p1, p2, p3]:
            idx0 = int(pmin * nsec * sampling_rate)
            idx1 = int(pmax * nsec * sampling_rate)
            signal[idx0:idx1] = np.nan
    return Trace(data=signal, header=stats)




import inspect
import os
import shutil
import tempfile
import unittest
import warnings

import numpy as np
import pandas as pd
import xarray as xr
from obspy import Trace, UTCDateTime
from zizou.cpxtrace import (
    Centroid, Inst_Band, Inst_Freq, Norm_Env, WindowCentroid
)
from zizou.data import DataSource
from zizou.util import test_signal, xarray2hdf5

from zizou.cpxtrace import Inst_Band, Inst_Freq, Norm_Env, Centroid, WindowCentroid


@unittest.skip("Tested code needs re-write")
class TestCpxtrace(unittest.TestCase):

    def setUp(self):
        # set the path to test data
        self.data_dir = os.path.join(os.path.dirname(os.path.abspath(
            inspect.getfile(inspect.currentframe()))), "data")
        self.tr1 = self.test_signal()
        self.tr2 = self.test_signal(gaps=True)

    def test_signal(self, gaps=False):
        nsec = 6400.
        sampling_rate = 100.
        signal = -1e-9 * np.ones(int(nsec * sampling_rate))
        if gaps:
            idx0 = int(500 * sampling_rate)
            idx1 = int(1200 * sampling_rate)
            signal[idx0:idx1] = np.nan

        stats = {'network': 'NZ', 'station': 'BLUB', 'location': '',
                 'channel': 'HHZ', 'npts': len(signal),
                 'sampling_rate': sampling_rate,
                 'mseed': {'dataquality': 'D'}}
        stats['starttime'] = UTCDateTime()
        stats['endtime'] = stats['starttime'] + nsec
        tr = Trace(data=signal, header=stats)
        return tr

    def test_with_gaps(self):
        instband = Inst_Band(interval=600., fil_coef=(-1.0, 0.0, 1.0))
        df_instB = instband.compute(self.tr1)
        np.testing.assert_almost_equal(
            df_instB.values,
            np.r_[np.linspace(0.00125562, 0.00125562, 10)].reshape((10, 1)),
            8
        )
        df_instB_gaps = instband.compute(self.tr2)
        np.testing.assert_almost_equal(
            df_instB_gaps.values,
            np.r_[np.nan, np.nan,
                  np.linspace(0.00125562, 0.00125562, 8)
                  ].reshape((10, 1)),
            8
        )

        instfreq = Inst_Freq(interval=600., fil_coef=(-1.0, 0.0, 1.0))
        df_instF = instfreq.compute(self.tr1)
        np.testing.assert_almost_equal(
            df_instF.values,
            np.r_[np.linspace(0.00141813, 0.00141813, 10)].reshape((10, 1)),
            8
        )
        df_instF_gaps = instfreq.compute(self.tr2)
        np.testing.assert_almost_equal(
            df_instF_gaps.values,
            np.r_[np.nan, np.nan,
                  np.linspace(0.00141813, 0.00141813, 8)
                  ].reshape((10, 1)),
            8
        )
        normenv = Norm_Env(window=10, interval=600., fil_coef=(-1.0, 0.0, 1.0))
        df_normE = normenv.compute(self.tr1)
        np.testing.assert_almost_equal(
            df_normE.values,
            np.r_[np.linspace(8.59312621e-12, 8.59312621e-12, 10)
                  ].reshape((10, 1)),
            8
        )

        df_normE_gaps = normenv.compute(self.tr2)
        np.testing.assert_almost_equal(
            df_normE_gaps.values,
            np.r_[np.nan, np.nan,
                  np.linspace(8.59312621e-12, 8.59312621e-12, 8)
                  ].reshape((10, 1)),
            8
        )

    def test_inst_band(self):
        """
        Test instantaneous bandwidth method
        """
        stream = 'WIZ.10.HHZ.NZ'
        ds = DataSource(source='fdsn')

        startdate = UTCDateTime('2019-07-14')
        enddate = UTCDateTime('2019-07-15')

        site, loc, comp, net = str.split(stream, '.')
        instband = Inst_Band(interval=600., fil_coef=(-1.0, 0.0, 1.0))
        for tr in ds.get_waveforms(net, site, loc, comp,
                                   startdate, enddate,
                                   fill_value='interpolate'):
            omega = instband.compute(tr)

        omega_test = pd.read_csv(os.path.join(self.data_dir,
                                              'WIZ_instband.csv'),
                                 index_col=0, parse_dates=True)

        pd.testing.assert_frame_equal(omega, omega_test)

    def test_inst_freq(self):
        """
        Test instantaneous frequency method
        """
        stream = 'WIZ.10.HHZ.NZ'
        ds = DataSource(source='fdsn')

        startdate = UTCDateTime('2019-07-14')
        enddate = UTCDateTime('2019-07-15')

        site, loc, comp, net = str.split(stream, '.')
        instfreq = Inst_Freq(interval=600., fil_coef=(-1.0, 0.0, 1.0))

        for tr in ds.get_waveforms(net, site, loc, comp,
                                   startdate, enddate,
                                   fill_value='interpolate'):
            sigma = instfreq.compute(tr)

        sigma_test = pd.read_csv(os.path.join(self.data_dir,
                                              'WIZ_instfreq.csv'),
                                 index_col=0, parse_dates=True)

        pd.testing.assert_frame_equal(sigma, sigma_test)

    def test_norm_env(self):
        """
        Test Normalize envelope method
        """
        stream = 'WIZ.10.HHZ.NZ'
        ds = DataSource(source='fdsn')

        startdate = UTCDateTime('2019-08-14T00:00:01')
        enddate = startdate + 86400.

        site, loc, comp, net = str.split(stream, '.')
        normenv = Norm_Env(window=10, interval=600., fil_coef=(-1.0, 0.0, 1.0))

        for tr in ds.get_waveforms(net, site, loc, comp,
                                   startdate, enddate,
                                   fill_value='interpolate'):
            norm_env = normenv.compute(tr)

        norm_env_test = pd.read_csv(os.path.join(self.data_dir,
                                                 'WIZ_norm_env.csv'),
                                    index_col=0, parse_dates=True)

        pd.testing.assert_frame_equal(norm_env, norm_env_test)

    @unittest.skip('Slow test')
    def test_centroid(self):
        """
        Test instantaneous frequency method
        """
        stream = 'WIZ.10.HHZ.NZ'
        ds = DataSource(source='fdsn')

        startdate = UTCDateTime('2019-08-14')
        enddate = startdate + 1000.

        site, loc, comp, net = str.split(stream, '.')

        for tr in ds.get_waveforms(net, site, loc, comp,
                                   startdate, enddate,
                                   attach_response=True):
            centro = Centroid.compute(tr)

        centro_test = pd.read_csv(os.path.join(self.data_dir,
                                               'WIZ_centroid.csv'),
                                  index_col=0, parse_dates=True)

        centro_test.columns = ['']

        pd.testing.assert_frame_equal(centro, centro_test)

@unittest.skip("Tested code needs re-write")
class WindowCentroidTestCase(unittest.TestCase):

    def setUp(self):
        self.data_dir = os.path.join(os.path.dirname(os.path.abspath(
            inspect.getfile(inspect.currentframe()))), "data")
        stream = 'WIZ.10.HHZ.NZ'
        ds = DataSource(source='fdsn')
        # dates are inclusive, and complete days
        startdate = UTCDateTime('2019-08-14')
        enddate = UTCDateTime('2019-08-15')

        site, loc, comp, net = str.split(stream, '.')

        gen = ds.get_waveforms(net, site, loc, comp,
                               startdate, enddate,
                               fill_value='interpolate')
        self.trace = next(gen)

    def test_window_centroid(self):
        feat = WindowCentroid(interval=600)
        df = feat.compute(self.trace)
        df_test = pd.read_csv(os.path.join(
            self.data_dir, 'window_centroid_test.csv'),
            index_col=0, parse_dates=True)
        pd.testing.assert_frame_equal(df, df_test)

if __name__ == '__main__':
    unittest.main()



from datetime import datetime

import matplotlib.pyplot as plt
from cftime import date2num, num2date
import datashader as ds
from matplotlib.colors import Normalize
from matplotlib import cm
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

def plot_ssam_mpl(xdf, log=False, axes=None, dbscale=False, cmap=cm.viridis,
                  clip=[0.0, 1.0], figsize=(12, 6)):
    """
    Plot spectrogram data using matplotlib.

    :param xdf: An xarray dataset as returned by :class:`zizou.ssam.SSAM`
    :type xdf: :class:`xarray.Dataset`
    :param log: Logarithmic frequency axis if True, linear frequency axis
                otherwise.
    :type log: bool
    :param axes: Plot into given axes.
    :type axes: :class:`matplotlib.axes.Axes`
    :param dbscale: If True 10 * log10 of color values is taken, if False the
                    sqrt is taken.
    :type dbscale: bool
    :param cmap: Specify a custom colormap instance.
    :type cmap: :class:`matplotlib.colors.Colormap`
    :param clip: adjust colormap to clip at lower and/or upper end. The given
                 percentages of the amplitude range (linear or logarithmic depending
                 on option `dbscale`) are clipped.
    :type clip: [float, float]
    :param figsize: Horizontal and vertical dimension 
                    of :class:`matplotlib.figure.Figure`
    :type figsize: [float, float]
    
    >>> import numpy as np
    >>> from zizou.ssam import SSAM, test_signal
    >>> from zizou.visualise import plot_ssam_mpl
    >>> ft = SSAM(interval=60, per_lap=.9, smooth=10, \
                  frequencies=np.arange(0., 25.1, .1), \
                  timestamp='start')
    >>> _ = ft.compute(test_signal())
    >>> fig = plot_ssam_mpl(ft.feature)

    """
    spec = xdf.data
    freq = xdf.frequency.data
    time = xdf.datetime.data
    time = pd.to_datetime(time).to_pydatetime()
    
    # db scale and remove zero/offset for amplitude
    if dbscale:
        spec = 10 * np.log10(spec[1:, :])
    else:
        spec = np.sqrt(spec[1:, :])
    freq = freq[1:]

    vmin, vmax = clip
    _range = float(spec.max() - spec.min())
    vmin = spec.min() + vmin * _range
    vmax = spec.min() + vmax * _range
    norm = Normalize(vmin, vmax, clip=True)
    # ignore warnings due to NaNs
    with np.errstate(invalid='ignore'):
        spec = np.where(spec < vmin, vmin, spec)
        spec = np.where(spec > vmax, vmax, spec)

    if not axes:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
    else:
        ax = axes

    # calculate half bin width
    halfbin_time = (time[1] - time[0]) / 2.0
    halfbin_freq = (freq[1] - freq[0]) / 2.0

    # argument None is not allowed for kwargs on matplotlib python 3.3
    kwargs = dict(cmap=cmap)

    if log:
        # pcolor expects one bin more at the right end
        freq = np.concatenate((freq, [freq[-1] + 2 * halfbin_freq]))
        time = np.concatenate((time, [time[-1] + 2 * halfbin_time]))
        # center bin
        time -= halfbin_time
        freq -= halfbin_freq
        # Log scaling for frequency values (y-axis)
        ax.set_yscale('log')
        # Plot times
        ps = ax.pcolormesh(time, freq, spec, norm=norm, **kwargs)
    else:
        # this method is much much faster!
        spec = np.flipud(spec)
        # center bin
        extent = (time[0] - halfbin_time, time[-1] + halfbin_time,
                  freq[0] - halfbin_freq, freq[-1] + halfbin_freq)
        ps = ax.imshow(spec, interpolation="nearest", extent=extent, **kwargs)

    # set correct way of axis, whitespace before and after with window
    # length
    ax.axis('tight')
    ax.grid(False)
    fig.colorbar(ps)

    if axes:
        return ax

    ax.set_ylabel('Frequency [Hz]')
    x_ticks = np.datetime64(xdf.attrs['starttime']) + ax.get_xticks().astype('timedelta64[s]')
    x_tick_labels = x_ticks.astype('datetime64[h]')
    ax.set_xticklabels(x_tick_labels, rotation=45)

    return fig


def plot_ssam_plotly(xdf, log=False, dbscale=False, cmap='Ice_r',
                     clip=[0.0, 1.0], new_fig=True, canvas_dim=(400,400)):
    """
    Plot spectrogram data using matplotlib.

    :param data: An xarray dataset as returned by :class:`zizou.ssam.SSAM`
    :type data: :class:`xarray.Dataset`
    :param log: Logarithmic frequency axis if True, linear frequency axis
                otherwise.
    :type log: bool
    :param dbscale: If True 10 * log10 of color values is taken, if False the
                    sqrt is taken.
    :type dbscale: bool
    :param cmap: Specify a custom colormap instance from
                 `plotly.colors.sequential`.
    :type cmap: list
    :param clip: adjust colormap to clip at lower and/or upper end. The given
                 percentages of the amplitude range (linear or logarithmic
                 depending on option `dbscale`) are clipped.
    :type clip: [float, float]
    :param new_fig: Whether to return a figure or just a graph object.
    :type new_flag: boolean
    :param max_tpoints: Maximum number of time intervals. If the datasets
                        exceeds that number it will be resampled
                        to max_tpoints.
    :type max_tpoints: int

    >>> import numpy as np
    >>> from zizou.ssam import SSAM, test_signal
    >>> from zizou.visualise import plot_ssam_plotly
    >>> ft = SSAM(interval=60, per_lap=.9, smooth=10, \
                  frequencies=np.arange(0., 25.1, .1), \
                  timestamp='start')
    >>> _ = ft.compute(test_signal())
    >>> fig = plot_ssam_plotly(ft.feature)
    """
    vmin, vmax = clip
    _range = float(xdf.max() - xdf.min())
    vmin = xdf.min() + vmin * _range
    vmax = xdf.min() + vmax * _range
    
    with np.errstate(invalid='ignore'):
        xdf = xdf.where(xdf > vmin, vmin)
        xdf = xdf.where(xdf < vmax, vmax)

    if dbscale:
        xdf = 10*np.log10(xdf)

    freq_dim = xdf.dims[0]
    if xdf[freq_dim].data[0] == 0:
        xdf[{freq_dim:0}] = np.ones(xdf.shape[1])*np.nan

    freq = xdf[freq_dim]
    dates = xdf['datetime']
    spec = xdf.data
    if canvas_dim is not None:
        dates = date2num(xdf.datetime.data.astype('datetime64[us]').astype(datetime),
                         units='hours since 1970-01-01 00:00:00.0',
                         calendar='gregorian')
        xdf = xdf.assign_coords({'datetime': dates})
        cvs = ds.Canvas(plot_width=canvas_dim[0],
                        plot_height=canvas_dim[1])
        agg = cvs.raster(source=xdf)
        freq, d, spec = agg.coords[freq_dim], agg.coords['datetime'], agg.data
        dates = num2date(d, units='hours since 1970-01-01 00:00:00.0', calendar='gregorian')

    if not new_fig: 
        return go.Heatmap(
            z=spec,
            x=dates,
            y=freq,
            colorscale=cmap)

    fig = go.Figure(data=go.Heatmap(
            z=spec,
            x=dates,
            y=freq,
            colorscale=cmap))

    fig.update_layout(
        xaxis_nticks=20,
        yaxis_title='Frequency [Hz]'
    )

    if log:
        fig.update_yaxes(type='log')

    return fig


def get_trace(data, colour, kws_ssam):
    """
    Return a plotly graphics object from a feature.

    :param data: Feature data to plot.
    :type data: :class:`xarray.DataArray`
    :param colour: Colour for line plots.
    :type colour: str 
    :returns: A 1-D line plot or a 2D heatmap.
    :rtype: :class:`plotly.graph_objs.Scatter` or
            :class:`plotly.graph_objs.Figure`
    """
    if data.name == 'ssam':
        return plot_ssam_plotly(data, new_fig=False,
                                dbscale=True, **kws_ssam)
    elif data.name == 'filterbank':
         return plot_ssam_plotly(data, new_fig=False,
                                 dbscale=True, canvas_dim=None,
                                 **kws_ssam)
    elif data.name == 'sonogram':
         return plot_ssam_plotly(data, new_fig=False,
                                 dbscale=False, canvas_dim=None,
                                 **kws_ssam)
    else:
        # data = data.fillna(data.mean())
        dates = data.datetime.data
        dates = pd.to_datetime(dates).to_pydatetime()
        featureData = data.to_dataframe()[data.name].interpolate('index').values
        return go.Scatter(x=dates,
                          y=featureData, opacity=0.8,
                          line_color=colour,
                          name='<b>{}</b>'.format(str(data.name)))


def multi_feature_plot(feature1, feature2, equal_yrange=False,
                       rangeslider=False, log=False, kws_ssam=None):
    """
    Return a combined time-series plot of two features.

    :param feature1: First feature to plot with y-scale on
                     the left.
    :type feature1: str
    :param feature2: Second feature to plot with y-scale on
                     the right.
    :type feature2: str
    :param equal_yrange: If True align yscales.
    :type equal_yrange: bool
    :param rangeslider: If True show secondary panel for zooming.
    :type rangeslider: bool
    :param kws_ssam: Keywords to be passed to
                     :class:`zizou.visualise.plot_ssam_plotly`
    :type kws_ssam: dict
    :returns: plotly figure that can be displayed in a Jupyter
              notebook or in a dash app.
    :rtype: :class:`plotly.graph_objs.Figure`
    """
    if kws_ssam is None:
        kws_ssam = dict()
    
    if feature2.name in ['ssam', 'sonogram', 'filterbank']:
        # Always plot 2D features first
        feature_tmp = feature1
        feature1 = feature2
        feature2 = feature_tmp

    fig = make_subplots(specs=[[{"secondary_y": True}]])

    trace = get_trace(feature1, 'deepskyblue', kws_ssam)
    fig.add_trace(trace, secondary_y=False)

    fig.update_yaxes(tickfont=dict(color='deepskyblue'),
                     title_font=dict(color='deepskyblue'))
    if feature2.name not in ['ssam', 'sonogram', 'filterbank']:
        if feature1.name in ['ssam', 'sonogram', 'filterbank']:
            trace = get_trace(feature2, 'darkred', kws_ssam)
            fig.add_trace(trace,
                          secondary_y=True)
            fig.update_yaxes(secondary_y=True,
                             tickfont=dict(color='darkred'),
                             title_font=dict(color='darkred'))
            fig.update_yaxes(tickfont=dict(color='#4d4db3'),
                             title_font=dict(color='#4d4db3'),
                             secondary_y=False)
        else:
            trace  = get_trace(feature2, 'dimgray', kws_ssam)
            fig.add_trace(trace,
                          secondary_y=True)
            fig.update_yaxes(secondary_y=True,
                             tickfont=dict(color='dimgray'),
                             title_font=dict(color='dimgray'))

    fig.update_layout(
        title={
            'text': "Features time series",
            'y': 0.9,
            'x': 0.5,
            'xanchor': 'center',
            'yanchor': 'top'},
        height=500,
        xaxis_rangeslider_visible=rangeslider)

    # Set y-axes titles
    fig.update_yaxes(title_text='<b>{}</b>'.format(str(feature1.name)), secondary_y=False)
    fig.update_yaxes(title_text='<b>{}</b>'.format(str(feature2.name)), secondary_y=True)
    
    if equal_yrange:
        if feature1.name in ['sonogram', 'filterbank', 'ssam']:
            # Smallest positive number
            y = fig['data'][0].y
            y_min = y[y>0][0] 
            y_max = fig['data'][0].y.max()
            y_range = np.array([y_min, y_max])
            if log:
                y_range = np.log10(y_range)
            fig.update_yaxes(range=y_range)
        else:
            y_min = min(float(feature1.min()),
                        float(feature2.min()))
            y_max = max(float(feature1.max()),
                        float(feature2.max()))
            if y_max > y_min:
                fig.update_yaxes(range=[y_min, y_max])

    if log:
        fig.update_yaxes(type='log')

    return fig


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import datetime
import inspect
import tempfile
import unittest
from unittest import TestCase

import numpy as np
import xarray as xr
import os
import pandas as pd

from obspy import UTCDateTime
from zizou import get_data
from zizou.pca import PCA, runPca, runPcaCore


@unittest.skip("Tested code needs re-write")
class PrincipalComponentTestCase(unittest.TestCase):

    def setUp(self):
        filename = inspect.getfile(inspect.currentframe())
        filedir = os.path.dirname(os.path.abspath(filename))
        self.testdata_dir = os.path.join(filedir, "data")

        class DummyFQTrain:
            def __init__(self):
                rng = np.random.RandomState(42)
                self.x, self.y = rng.multivariate_normal([10, 3],
                                                         cov=[[2, 1], [1, 2]],
                                                         size=10000).T
                self.times = pd.date_range('2021-05-25', freq='10min', periods=10000)

            def __call__(self, feature, stack_length=None):
                if feature == 'rsam':
                    return xr.DataArray(data=self.x,
                                        coords=[self.times],
                                        dims=['datetime'],
                                        name='rsam',
                                        )
                elif feature == 'dsar':
                    return xr.DataArray(data=self.x,
                                        coords=[self.times],
                                        dims=['datetime'],
                                        name='dsar')

        self.dfq_train = DummyFQTrain()

        class DummyFQTransform:
            def __init__(self):
                self.x = np.array([1., np.nan, 3., np.nan])
                self.y = np.array([1., np.nan, 3.])

                self.starttime = datetime.datetime(2020, 1, 1, 0, 0, 0)
                self.nantime = datetime.datetime(2020, 1, 1, 0, 10, 0)
                self.endtime = datetime.datetime(2020, 1, 1, 0, 20, 0)
                self.extraTime = datetime.datetime(2020, 1, 1, 0, 30, 0)
                self.times = [self.starttime, self.nantime, self.endtime]
                self.site = 'RIZ'

            def __call__(self, feature, stack_length=None):
                if feature == 'rsam':
                    return xr.DataArray(data=self.x,
                                        coords=[self.times+[self.extraTime]],
                                        dims=['datetime'],
                                        name="x")
                elif feature == 'dsar':
                    return xr.DataArray(data=self.y,
                                        coords=[self.times],
                                        dims=['datetime'],
                                        name='y')

        self.dfq_transform = DummyFQTransform()
        self.testfile = tempfile.mktemp()

    def tearDown(self):
        try:
            os.unlink(self.testfile)
        except FileNotFoundError:
            pass

    def test_normalise_and_fill(self):
        print("test_normalise_and_fill")
        pca = PCA(self.testfile,
                  n_components=2,
                  pca_features=['rsam', 'dsar'])

        feat1 = self.dfq_train(pca.pca_features[0]) #stack_dict[pca.pca_features[0]])

        mn, norm_feat_fill_nan, std = pca.normaliseAndFill(feat1)
        self.assertEqual(mn, 9.9900799153757021)
        self.assertEqual(std, 1.4216981693126631)
        array_sum = np.sum(norm_feat_fill_nan)
        array_has_nan = np.isnan(array_sum)
        self.assertFalse(array_has_nan)

        feat2 = self.dfq_transform(pca.pca_features[0], stack_dict[pca.pca_features[0]])
        mn, norm_feat_fill_nan, std = pca.normaliseAndFill(feat2)
        self.assertEqual(mn, 2.0)
        self.assertEqual(std, 1.0)

    def test_pca_train(self):
        print("test_pca_train")
        pca = PCA(self.testfile, n_components=2,
                  pca_features=['rsam', 'dsar'])
        pca.train(self.dfq_train)
        tval = 1/np.sqrt(2)
        np.testing.assert_array_almost_equal(np.abs(pca.pca_model.components_),
                                             np.array([[tval, tval],
                                                       [tval, tval]]), 8)
        self.assertEqual(np.abs(np.sign(pca.pca_model.components_[0]).sum()),
                         2)
        self.assertEqual(np.abs(np.sign(pca.pca_model.components_[1]).sum()),
                         0)

    def test_pca_transform(self):
        print("test_pca_transform")
        pca = PCA(self.testfile, n_components=2,
                  pca_features=['rsam', 'dsar'])
        pca.train(self.dfq_train)
        xdf = pca.infer(self.dfq_transform)
        self.assertTrue(np.all(np.isclose(xdf.pca.loc[:, 'pc1'].values, 0)))

        # test caching
        pca1 = PCA(self.testfile, n_components=2,
                   pca_features=['rsam', 'dsar'])
        pca1.train('dummy')
        xdf1 = pca1.infer(self.dfq_transform)
        self.assertTrue(np.all(np.isclose(xdf1.pca.loc[:, 'pc1'].values, 0)))
        tval = np.sqrt(2)
        np.testing.assert_array_almost_equal(xdf1.pca.loc[:, 'pc0'].values,
                                             np.array([tval, 0, -tval]), 8)

    # def test_pcaRun(self):
    #
    #     runPca()

    # def test_runPcaCore(self):
    #     print("test_runPcaCore")
    #     start = UTCDateTime(2021, 3, 13, 6, 0, 0)
    #     end = round_times(start, 600)
    #     start = end-7200
    #
    #     config_file = os.path.join(self.testdata_dir, "config_tests.ini")
    #     meta_file = get_data('package_data/metadata.json')
    #
    #     outDir = tempfile.gettempdir()
    #     runPcaCore(config_file=config_file, metadata_file=meta_file,
    #                end_time=end.datetime, start_time=start.datetime,
    #                out_dir=outDir)


if __name__ == '__main__':
    unittest.main()



import math as M

import numpy as np
import obspy
import pandas as pd
import xarray as xr
from zizou import FeatureBaseClass
from zizou.util import window_array
import yaml


class SSAM(FeatureBaseClass):
    """
    SSAM is basically the same as computing a spectrogram. Data are split into
    interval length segments and the spectrum of each section is
    computed. Before computing the spectrum, the mean is removed and
    a hanning window is applied to each segment. The percentage of
    overlap of each segment can be specified with per_lap and Welch-type
    smoothing can be applied.

    :param interval: Segment length in seconds.
    :type interval: int
    :param per_lap: If the value is less than 1 it is treated as the percentage
                    of segment overlap; else it is the step size in sample points
    :type per_lap: float or int
    :param scale_by_freq: Divide by the sampling frequency so that density
                          function has units of dB/Hz and can be integrated
                          by the frequency values.
    :type scale_by_freq: boolean
    :param smooth: The number of windows over which to
                   average the spectrogram.
    :type smooth: int
    :param frequencies: Frequencies at which to return the spectrogram. This
                        uses linear interpolation to compute the spectrogram
                        at the given frequencies from the original spectrogram.
    :type frequencies: :class:`numpy.ndarray`
    :param timestamp: Can be either 'center' or 'start'. If 'center', the
                      timestamp of each spectrum is the center of the windows
                      from which the spectrum was computed. If 'start' it is
                      the timestamp of the first sample of the first window.
    :type timestamp: str
    :param resample_int: Interval to upsample the dataset using linear
                         interpolation over the first interval and then
                         downsample using the mean over the second interval.
    :type resample_int: (str, str)
    :param configfile: Configuration as an .ini file or a dictionary.
    :type configfile: str or dict

    >>> from zizou.ssam import SSAM, test_signal
    >>> import numpy as np
    >>> tr = test_signal(gaps=True)
    >>> s = SSAM(interval=60, per_lap=.9, smooth=10,
                 frequencies=np.arange(0, 25.1, .1))

    """

    features = ['ssam', 'filterbank', 'sonogram']

    def __init__(self, interval=10, per_lap=0, scale_by_freq=True,
                 smooth=None, frequencies=np.linspace(0, 25, 251),
                 fbfilters=16, nhob=8, timestamp='center',
                 resample_int=None, configfile=None):
        super(SSAM, self).__init__()
        self.interval = interval
        self.per_lap = per_lap
        self.scale_by_freq = scale_by_freq
        self.smooth = smooth
        self.frequencies = frequencies
        self.timestamp = timestamp
        self.resample_int = resample_int
        self.fbfilters = fbfilters
        self.nhob = nhob 

        self.feature = None

        if configfile is not None:
            try:
                with open(configfile, 'r') as fh:
                    c = yaml.safe_load(fh)
            except OSError:
                c = yaml.safe_load(configfile)
            cs = c.get('ssam')
            if cs is not None:
                self.interval = cs.get('interval', interval)
                self.per_lap = cs.get('per_lap',per_lap)
                self.scale_by_freq = cs.get('scale_by_freq', scale_by_freq)
                self.smooth = cs.get('smooth', smooth)
                self.timestamp = cs.get('timestamp', timestamp)
                self.nhob = cs.get('nhob', nhob)
                freqs = cs.get('frequencies', frequencies)
                if freqs is not None:
                    self.frequencies = np.arange(freqs['start'],
                                                 freqs['end'],
                                                 freqs['step'])

                self.fbfilters = cs.get('fbfilters', fbfilters)
                ri = cs.get('resample_int', resample_int)
                if ri is not None:
                    self.resample_int = (ri['upsample'], ri['downsample'])

    @staticmethod
    def _nearest_pow_2(x):
        """
        Find power of two nearest to x

        >>> _nearest_pow_2(3)
        2.0
        >>> _nearest_pow_2(15)
        16.0

        :param x: Number
        :type x: float
        :return: Nearest power of 2 to x
        :rtype: Int
        """
        a = M.pow(2, M.ceil(np.log2(x)))
        b = M.pow(2, M.floor(np.log2(x)))
        if abs(a - x) < abs(b - x):
            return a
        else:
            return b

    def _average(self, spec, t):
        """
        Average spectrogram over nwin windows to produce a smoother
        spectrogram. If used with overlapping windows it is known
        as the Welch algorithm

        :param spec: Input spectrogram
        :type spec: :class:`numpy.ndarray`
        :param t: Time axis of spectrogram.
        :type t: :class:`numpy.ndarray`
        :param nwin: Number of windows over which to average
        :type nwin: int
        :return: smoothed_spec, new-time
        :rtype: tuple of :class:`~numpy.ndarray`
        """
        start = 0
        end = 0
        ntimes = spec.shape[1]
        av_mat = []
        nt = []
        while end < ntimes:
            av_arr = np.zeros(spec.shape[1])
            end = start + self.smooth
            if end > ntimes:
                end = ntimes
            av_arr[start:end] = 1.
            av_mat.append(av_arr)
            if self.timestamp == 'center':
                nt.append(t[start:end:self.smooth - 1].mean())
            elif self.timestamp == 'start':
                nt.append(t[start])
            start = end

        av_mat = np.array(av_mat).T
        nt = np.array(nt)
        # find all columns with NaNs
        colidx = np.unique(np.where(np.isnan(spec))[1])
        # set corresponding rows in averaging matrix to NaN
        av_mat[colidx, :] = 0
        factor = np.sum(av_mat, axis=0)
        av_mat = np.divide(av_mat, factor, where=av_mat > 0)
        mspec = np.ma.masked_invalid(spec)
        mav_mat = np.ma.masked_invalid(av_mat)
        nspec = np.ma.dot(mspec, mav_mat)
        return nspec.data, nt

    @staticmethod
    def _resample(xdf, resample_int=('1min', '10min')):
        """
        Resample spectrograms to the desired sampling interval using xarray
        functionality.

        :param xdf: An xarray dataset as returned by :class:`zizou.ssam.SSAM`
        :type xdf: :class:`xarray.Dataset`
        :param resample_int: Intervals to upsample the dataset using linear
                             interpolation over the first interval and then
                             downsample using the mean over the second
                             interval.
        :type resample_int: (str, str)
        :return: Stacked dataset.
        :rtype: :class:`~xarray.Dataset`
        """
        upsample_int, downsample_int = resample_int
        attrs = xdf.attrs
        if upsample_int is not None:
            xdf = xdf.resample(datetime=upsample_int).interpolate('linear')
        if downsample_int is not None:
            xdf = xdf.resample(datetime=downsample_int).mean()
        # Resample transposes the dimensions so we need to revert this
        xdf['ssam'] = xdf.ssam.transpose('frequency', 'datetime')
        xdf['filterbank'] = xdf.filterbank.transpose('fbfrequency', 'datetime')
        xdf['sonogram'] = xdf.sonogram.transpose('sonofrequency', 'datetime')
        xdf.attrs = attrs
        return xdf

    @staticmethod
    def filterbank(nfilters, nfft, fs):
        """
        Computes the integral in an array of bandpass filters that
        separates the input signal into multiple components.
        Each component consists of a single frequency sub-band
        of the original signal.
        
        :param nfilters: Number of filters in the filter bank.
        :type nfitlers: int
        :param nfft: Number of samples
        :type nfft: int
        :param fs: Sampling frquency
        :type fs: float

        """
        fl = np.floor(fs) / np.floor(nfft)
        fh = np.floor(fs / 2)
        freqs = np.logspace(
            np.log10(fl), np.log10(fh), nfilters + 2
        )
        bins = np.floor((nfft + 1) * freqs / fs)
        fbank = np.zeros((nfilters, int(np.floor(nfft / 2 + 1))))
        for m in range(1, nfilters + 1):
            f_m_minus = int(bins[m - 1])   # left
            f_m = int(bins[m])             # center
            f_m_plus = int(bins[m + 1])    # right

            for k in range(f_m_minus, f_m):
                fbank[m - 1, k] = (k - bins[m - 1]) / (bins[m] - bins[m - 1])
            for k in range(f_m, f_m_plus):
                fbank[m - 1, k] = (bins[m + 1] - k) / (bins[m + 1] - bins[m])

        return freqs[1:-1], fbank

    @staticmethod
    def sonogram(data, fs, fc1, nofb=8):
        """
        Computes the sonogram of the given data which can be windowed or not.
        The sonogram is determined by the power in half octave bands of the given
        data.

        If data are windowed the analytic signal and the envelope of each window
        is returned.

        :param data: Data to make envelope of.
        :type data: :class:`~numpy.ndarray`
        :param fs: Sampling frequency in Hz.
        :param fc1: Center frequency of lowest half octave band.
        :param nofb: Number of half octave bands.
        :param no_win: Number of data windows.
        :return: Central frequencies, Half octave bands.
        """
        fc = float(fc1) * 1.5**np.arange(nofb)
        fmin = fc / np.sqrt(5. / 3.)
        fmax = fc * np.sqrt(5. / 3.)
        z_tot = np.sum(np.abs(data)**2, axis=0)
        nfft = data.shape[0]
        start = np.around(fmin * nfft / fs, 0).astype(np.int_) - 1
        end = np.around(fmax * nfft / fs, 0).astype(np.int_)
        z = np.zeros([nofb, data.shape[1]])
        for i in range(nofb):
            z[i, :] = np.sum(np.abs(data[start[i]:end[i], :])**2, axis=0)
        hob = np.log(z / z_tot[np.newaxis, :])
        return fc, hob

    def compute(self, trace):
        """
        Compute SSAM for the given trace.

        :param trace: The seismic waveform data.
        :type trace: :class:`obspy.Trace`
        :return: spectrogram, frequencies, times
        :rtype: tuple
        """
        if len(trace) < 1:
            msg = "Trace is empty."
            raise ValueError(msg)
        tr = trace.copy()
        x = tr.data
        npts = tr.stats.npts
        Fs = tr.stats.sampling_rate
        interval = min(self.interval*Fs, npts)

        NFFT = int(self._nearest_pow_2(interval))
        if NFFT > npts:
            NFFT = int(self._nearest_pow_2(npts / 8.0))

        if self.per_lap < 1.:
            noverlap = int(NFFT * float(self.per_lap))
        else:
            noverlap = NFFT - self.per_lap

        scaling_factor = 2.
        result, windowVals = window_array(x, NFFT, noverlap)
        result = np.fft.rfft(result, n=NFFT, axis=0)
        freqs = np.fft.rfftfreq(NFFT, 1 / Fs)

        result = np.conj(result) * result
        # Also include scaling factors for one-sided densities and dividing by
        # the sampling frequency, if desired. Scale everything, except the DC
        # component and the NFFT/2 component:
        slc = slice(1, -1, None)

        result[slc] *= scaling_factor

        # MATLAB divides by the sampling frequency so that density function
        # has units of dB/Hz and can be integrated by the plotted frequency
        # values. Perform the same scaling here.
        if self.scale_by_freq:
            result /= Fs
            # Scale the spectrum by the norm of the window to compensate for
            # windowing loss; see Bendat & Piersol Sec 11.5.2.
            result /= (np.abs(windowVals)**2).sum()
        else:
            # In this case, preserve power in the segment, not amplitude
            result /= np.abs(windowVals).sum()**2

        result = result.real
        steplength = NFFT - noverlap
        steps = np.arange(result.shape[1]) * steplength
        if self.timestamp == 'center':
            t = (steps + NFFT / 2) / Fs
        elif self.timestamp == 'start':
            t = steps / Fs
        # smooth the spectrogram using the Welch algorithm
        if self.smooth is not None:
            result, t = self._average(result, t)

        features = {}
        dates = [(trace.stats.starttime + _t).datetime for _t in t]

        fb_freqs, fb = self.filterbank(self.fbfilters, NFFT, Fs)
        result_fb = np.dot(fb, result)
        filterbank = xr.DataArray(result_fb,
                                  coords=[fb_freqs,
                                          pd.to_datetime(dates)],
                                  dims=['fbfrequency', 'datetime'])
        features['filterbank'] = filterbank
            
        sono_freq, sono_amp = self.sonogram(result, Fs, 0.68, self.nhob)
        sono = xr.DataArray(sono_amp,
                            coords=[sono_freq,
                                    pd.to_datetime(dates)],
                            dims=['sonofrequency', 'datetime'])
        features['sonogram'] = sono

        # Linear interpolation onto a new set of frequencies
        def myinterp(a):
            return np.interp(self.frequencies, freqs, a)
        result = np.apply_along_axis(myinterp, 0, result)
        freqs = self.frequencies

        ssam = xr.DataArray(
            result,
            coords=[freqs, pd.to_datetime(dates)],
            dims=['frequency', 'datetime']
        )
        features['ssam'] = ssam
        self.feature = xr.Dataset(features)
        self.feature.attrs['starttime'] = trace.stats.starttime.isoformat()
        self.feature.attrs['endtime'] = trace.stats.endtime.isoformat()
        self.feature.attrs['station'] = trace.stats.station

        if self.resample_int is not None:
            self.feature = self._resample(self.feature, self.resample_int)
        return self.feature



from datetime import datetime, timedelta
import inspect
import os
from subprocess import Popen

import pytest
import yaml

from zizou.data import FeatureRequest
from zizou.util import generate_test_data, xarray2hdf5

def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


tstart = datetime(2023,1,1,0,0,0)
ndays = 10

@pytest.fixture(scope='package')
def setup(tmp_path_factory):
    features1D = ['rsam',
                  'dsar',
                  'central_freq',
                  'predom_freq',
                  'bandwidth',
                  'rsam_energy_prop']
    features2D = [('sonogram', 'sonofrequency'),
                  ('ssam', 'frequency'),
                  ('filterbank', 'fbfrequency')]

    savedir = tmp_path_factory.mktemp('zizou_test_tmp', numbered=True)
    widir = os.path.join(savedir,
                         'features',
                         'Whakaari',
                         'WIZ',
                         'HHZ')
    doomdir = os.path.join(savedir,
                           'features',
                           'Mt Doom',
                           'MDR',
                           'BHZ')
    ruadir = os.path.join(savedir,
                         'features',
                         'Ruapehu',
                         'MAVZ',
                         'EHZ')
    mmdir = os.path.join(savedir,
                         'features',
                         'Misty Mountain',
                         'MMS',
                         'BHZ')
    for _dir in [widir, doomdir, ruadir, mmdir]:
        os.makedirs(_dir)
   # Generate some fake data
    for _f in features1D:
        feat = generate_test_data(tstart=tstart,
                                    feature_name=_f,
                                    ndays=ndays)
        for _dir in [widir, doomdir, ruadir]:
            xarray2hdf5(feat, _dir)
    for _n, _f in features2D:
        feat = generate_test_data(tstart=tstart,
                                  feature_name=_n,
                                  ndays=ndays,
                                  nfreqs=8,
                                  freq_name=_f,
                                  dim=2)
        for _dir in [widir, doomdir, ruadir]:
            xarray2hdf5(feat, _dir)

    alg = generate_test_data(tstart=tstart,
                             feature_name='autoencoder',
                             ndays=ndays,
                             nfreqs=5,
                             freq_name='cluster',
                             dim=2)
    xarray2hdf5(alg, doomdir)
    data_dir = os.path.join(os.path.dirname(os.path.abspath(
                            inspect.getfile(inspect.currentframe()))), "data")
    with open(os.path.join(data_dir, 'config_integration_test.yml')) as fh:
        cfg = yaml.safe_load(fh)
    cfg['default'] = {'outdir': os.path.join(savedir, 'features')}
    fn_cfg = os.path.join(savedir, 'tmp.yml')
    with open(fn_cfg, 'w') as configfile:
        yaml.dump(cfg, configfile)
    return savedir

@pytest.fixture(scope='module')
def setup_ac(setup):
    savedir = setup
    fq = FeatureRequest(rootdir=os.path.join(savedir, 'features'),
                        volcano='Whakaari',
                        site='WIZ',
                        channel='HHZ',
                        starttime=tstart,
                        endtime=tstart + timedelta(days=ndays))
    config = """ 
    autoencoder:
      layers: [2000,500,200,6]
      epochs: 5
      patience: 10
    """
    return savedir, fq, config


@pytest.fixture(scope='module')
def setup_orchestrate(tmp_path_factory):
    savedir = tmp_path_factory.mktemp('zizou_test_tmp', numbered=True)
    test_dir = os.path.dirname(os.path.abspath(
                            inspect.getfile(inspect.currentframe())))
    data_dir = os.path.join(test_dir, "data")
    with open(os.path.join(data_dir, 'config_integration_test.yml')) as fh:
        cfg = yaml.safe_load(fh)
    cfg['default']['sds_dir'] =  os.path.join(savedir, 'sds')
    feature_dir = os.path.join(savedir, 'features')
    os.makedirs(feature_dir, exist_ok=True)
    cfg['default']['staxml_dir'] = data_dir
    cfg['default']['outdir'] = feature_dir
    cfg['default']['logdir'] = str(savedir)
    server = Popen(["random_walk_model"])
    fn_cfg = os.path.join(savedir, 'tmp.yml')
    with open(fn_cfg, 'w') as configfile:
        yaml.dump(cfg, configfile)
    os.environ['SAMCONFIG'] = fn_cfg
    yield feature_dir
    server.terminate()
 


import importlib
from os import PathLike
from typing import Optional

from zizou.anomaly_base import AnomalyDetectionBaseClass
# from zizou.pca import PCA
from zizou.data import FeatureRequest
from zizou.feature_base import FeatureBaseClass
from zizou.rsam import RSAM
from zizou.spectral_features import SpectralFeatures
from zizou.ssam import SSAM

__all__ = [
    "AnomalyDetectionBaseClass",
    "FeatureRequest",
    "FeatureBaseClass",
    "RSAM",
    "SpectralFeatures",
    "SSAM",
    "get_data",
]


def get_data(filename: Optional[PathLike] = None) -> str:
    """Return path to zizou package.

    Parameters
    ----------
    filename : Pathlike, default None
        Append `filename` to returned path.

    Returns
    -------
    pkgdir_path

    """
    f = importlib.resources.files(__package__)
    return str(f) if filename is None else str(f / filename)



from datetime import datetime, timedelta
import shutil
import tempfile
import unittest

import os

import xarray as xr

from zizou import AnomalyDetectionBaseClass
from zizou.util import generate_test_data, xarray2hdf5
from zizou.data import FeatureRequest


class AnomalyBaseClassTestCase(unittest.TestCase):

    def setUp(self):
        self.savedir = tempfile.mkdtemp()
        self.rootdir = os.path.join(self.savedir,
                                    'Mt_Doom',
                                    'MDR',
                                    'HHZ')
        os.makedirs(self.rootdir)
        tstart = datetime(2023,1,1,0,0,0)
        self.fq = FeatureRequest(rootdir=self.savedir,
                                 volcano='Mt_Doom',
                                 site='MDR',
                                 channel='HHZ',
                                 starttime=tstart,
                                 endtime=tstart + timedelta(days=30))
        # Generate some fake data
        rsam = generate_test_data(tstart=tstart,
                                  feature_name='rsam',
                                  ndays=30)
        xarray2hdf5(rsam, self.rootdir)
        dsar = generate_test_data(tstart=tstart,
                                  feature_name='dsar',
                                  ndays=30)
        xarray2hdf5(dsar, self.rootdir)
        sonogram = generate_test_data(tstart=tstart, feature_name='sonogram',
                                      freq_name='sonofrequency', nfreqs=8,
                                      dim=2,
                                      ndays=30)
        xarray2hdf5(sonogram, self.rootdir)
        
    def tearDown(self):
        if os.path.isdir(self.savedir):
            shutil.rmtree(self.savedir)

    def test_get_features(self):
        transform_dict = {'rsam': 'log'}
        stack_dict = {'dsar': '2D',
                      'central_freq': '1H',
                      'predom_freq': '1H',
                      'variance': '1H',
                      'bandwidth': '1H'}
        ab = AnomalyDetectionBaseClass(['rsam', 'sonogram', 'dsar'], stacks=stack_dict,
                                       transforms=transform_dict)
        featurefile = os.path.join(self.savedir, 'pca_features.nc')
        feats = ab.get_features(self.fq, featurefile)
        self.assertEqual(feats.shape, (30*144, 10))
        # Run again to test reading features from file
        feats1 = ab.get_features(self.fq, featurefile)
        xr.testing.assert_allclose(feats, feats1)


if __name__ == '__main__':
    unittest.main()

        




import inspect
import os
import shutil
import tempfile
import unittest

import numpy as np
from obspy import Trace, UTCDateTime
import pandas as pd
import xarray as xr

from zizou.rsam import RSAM, EnergyExplainedByRSAM
from zizou.data import DataSource


class RSAMTestCase(unittest.TestCase):

    def setUp(self):
        self.tr1 = self.test_signal()
        self.tr2 = self.test_signal(gaps=True)
        self.tempdir1 = tempfile.mkdtemp()
        self.tempdir2 = tempfile.mkdtemp()
        self.tempdir3 = tempfile.mkdtemp()

    def tearDown(self):
        for _dir in [self.tempdir1, self.tempdir2, self.tempdir3]:
            if os.path.isdir(_dir):
                shutil.rmtree(_dir)

    def test_signal(self, gaps=False, starttime=None):
        nsec = 6400.
        sampling_rate = 100.
        signal = -1e-9 * np.ones(int(nsec * sampling_rate))
        if gaps:
            idx0 = int(500 * sampling_rate)
            idx1 = int(1200 * sampling_rate)
            signal[idx0:idx1] = np.nan

        stats = {'network': 'NZ', 'station': 'BLUB', 'location': '',
                 'channel': 'HHZ', 'npts': len(signal),
                 'sampling_rate': sampling_rate,
                 'mseed': {'dataquality': 'D'}}
        stats['starttime'] = UTCDateTime()
        if starttime is not None:
            stats['starttime'] = starttime
        stats['endtime'] = stats['starttime'] + nsec
        tr = Trace(data=signal, header=stats)
        return tr

    def test_with_gaps(self):
        r = RSAM(filtertype=None)
        df = r.compute(self.tr1)
        np.testing.assert_array_almost_equal(df['rsam'].values, np.ones(10), 2)
        df1 = r.compute(self.tr2)
        np.testing.assert_array_almost_equal(
            df1['rsam'].values, np.r_[np.nan, np.nan, np.ones(8)], 2
        )

    def test_append(self):
        """
        Test appending new data to existing output files.
        """

        starttime1 = UTCDateTime('2019-08-14T01:00:00')
        endtime1 = UTCDateTime('2019-08-14T01:20:00')
        starttime2 = endtime1
        endtime2 = UTCDateTime('2019-08-14T01:40:00')

        tr = self.test_signal(starttime=starttime1)
        tr1 = tr.copy().trim(starttime1, endtime1)
        tr2 = tr.copy().trim(starttime2, endtime2)
        tr2.data *= 2
        # hack to make test results consistent
        # necessary because time windows overlap
        tr1.data[-1] *= 2
        tr3 = tr1.__add__(tr2, method=1)
        r1 = RSAM(interval=600, filtertype=None)
        r2 = RSAM(interval=600, filtertype=None)
        r3 = RSAM(interval=600, filtertype=None)

        _ = r1.compute(tr1)
        r1.save(self.tempdir1)

        _ = r2.compute(tr2)
        r2.save(self.tempdir1)

        _ = r3.compute(tr3)
        r3.save(self.tempdir2)

        file1 = os.path.join(self.tempdir1, 'rsam.nc')
        file2 = os.path.join(self.tempdir2, 'rsam.nc')
        ds1 = xr.open_dataset(file1, group='original')
        ds2 = xr.open_dataset(file2, group='original')
        np.testing.assert_array_equal(ds1['datetime'],
                                      ds2['datetime'])
        np.testing.assert_array_equal(ds1['rsam'],
                                      ds2['rsam'])

        # test overlapping time windows
        tr4 = tr.copy().trim(starttime1, endtime2)
        r4 = RSAM(interval=600, filtertype=None)
        _ = r4.compute(tr4)
        r4.save(self.tempdir3)
        r2.save(self.tempdir3)
        file3 = os.path.join(self.tempdir3, 'rsam.nc')
        ds3 = xr.open_dataset(file3, group='original')
        np.testing.assert_array_equal(ds3['rsam'].values.astype(int),
                                      np.array([1, 1, 2, 2]))


class EnergyExplainedByRSAMTestCase(unittest.TestCase):

    def setUp(self):
        self.data_dir = os.path.join(os.path.dirname(os.path.abspath(
            inspect.getfile(inspect.currentframe()))), "data")
        stream = 'WIZ.10.HHZ.NZ'
        ds = DataSource(sources=('fdsn',), fill_value='interpolate')
        # dates are inclusive, and complete days
        startdate = UTCDateTime('2019-08-14')
        enddate = UTCDateTime('2019-08-15')

        site, loc, comp, net = str.split(stream, '.')

        gen = ds.get_waveforms(net, site, loc, comp,
                               startdate, enddate)
        self.trace = next(gen)

    def test_energy(self):
        feat = EnergyExplainedByRSAM(
            interval=600, filtertype='bp', filterfreq=(2, 5)
        )
        df = feat.compute(self.trace).to_dataframe()
        fname = os.path.join(self.data_dir, 'rsam_energy_test.csv')
        df_test = pd.read_csv(
            fname, index_col="datetime", parse_dates=True
        )
        pd.testing.assert_frame_equal(df, df_test)




from configparser import ConfigParser
from datetime import datetime
import logging
import logging.config

import pytest

from zizou.autoencoder import AutoEncoder
from zizou import get_data
from zizou import zizou_logging


logger = logging.getLogger(__name__)


@pytest.mark.slow
def test_fit(setup_ac):
    rootdir, fq, config = setup_ac
    ac = AutoEncoder(config)
    ac.fit(fq)
    assert float(ac.loss_) - 1 < 0.5


@pytest.mark.slow
def test_transform(setup_ac):
    savedir, fq, config = setup_ac
    starttime = datetime(2023,1,8,0,0,0)
    endtime = datetime(2023,1,9,0,0,0)
    fq.starttime = starttime
    fq.endtime = endtime
    ac = AutoEncoder(config)
    ed = ac.fit_transform(fq, cluster=False)
    assert ed['autoencoder'].shape == (6, 144)
 

@pytest.mark.slow
def test_cluster(setup_ac):
    savedir, fq, config = setup_ac
    starttime = datetime(2023,1,8,0,0,0)
    endtime = datetime(2023,1,9,0,0,0)
    fq.starttime = starttime
    fq.endtime = endtime
    ac = AutoEncoder(config)
    ed = ac.fit_transform(fq)
    assert ed['autoencoder'].shape == (5, 144)
 
@pytest.mark.slow
def test_early_stopping(setup_ac):
    savedir, fq, config = setup_ac
    starttime = datetime(2023,1,8,0,0,0)
    endtime = datetime(2023,1,9,0,0,0)
    fq.starttime = starttime
    fq.endtime = endtime
    ac = AutoEncoder(config)
    ed = ac.fit_transform(fq, cluster=False)
    assert ed['autoencoder'].shape == (6, 144)
    ac.save(savedir)
 


# import unittest
# import doctest
# import zizou.visualise

# def load_tests(tests):
#     tests.addTests(doctest.DocTestSuite(zizou.visualise))
#     return tests

# if __name__ == '__main__':
#     runner = unittest.TextTestRunner()
#     suite = unittest.TestSuite()
#     runner.run(load_tests(suite))




import logging
import logging.config

ERROR_LOG_FILENAME = ".zizou-processing-errors.log"

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {      
        "default": {  # The formatter name, it can be anything that I wish
            "format": "%(asctime)s:%(name)s:%(process)d:%(lineno)d " "%(levelname)s %(message)s",  #  What to add in the message
            "datefmt": "%Y-%m-%d %H:%M:%S",  # How to display dates
        },
        "json": {  # The formatter name
         "()": "pythonjsonlogger.jsonlogger.JsonFormatter",  # The class to instantiate!
            # Json is more complex, but easier to read, display all attributes!
            "format": """
                    asctime: %(asctime)s
                    created: %(created)f
                    filename: %(filename)s
                    funcName: %(funcName)s
                    levelname: %(levelname)s
                    levelno: %(levelno)s
                    lineno: %(lineno)d
                    message: %(message)s
                    module: %(module)s
                    msec: %(msecs)d
                    name: %(name)s
                    pathname: %(pathname)s
                    process: %(process)d
                    processName: %(processName)s
                    relativeCreated: %(relativeCreated)d
                    thread: %(thread)d
                    threadName: %(threadName)s
                    exc_info: %(exc_info)s
                """,
            "datefmt": "%Y-%m-%d %H:%M:%S",  # How to display dates
        },
    }, 
    "handlers": {
        "logfile": {  # The handler name
            "formatter": "json",  # Refer to the formatter defined above
            "level": "ERROR",  # FILTER: Only ERROR and CRITICAL logs
            "class": "logging.handlers.RotatingFileHandler",  # OUTPUT: Which class to use
            "filename": ERROR_LOG_FILENAME,  # Param for class above. Defines filename to use, load it from constant
            "backupCount": 2,  # Param for class above. Defines how many log files to keep as it grows
        }, 
        "simple": {  # The handler name
            "formatter": "default",  # Refer to the formatter defined above
            "class": "logging.StreamHandler",  # OUTPUT: Same as above, stream to console
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": { 
        "zizou": {  # The name of the logger, this SHOULD match your module!
            "level": "DEBUG",  # FILTER: only INFO logs onwards from "tryceratops" logger
            "handlers": [
                "simple",  # Refer the handler defined above
            ],
        },
    },
    "root": {
        "level": "ERROR",  # FILTER: only INFO logs onwards
        "handlers": [
            "logfile",  # Refer the handler defined above
        ]
    },
}

logging.config.dictConfig(LOGGING_CONFIG)


import os
import stat

import pandas as pd
from zizou.util import xarray2hdf5
import xarray as xr
from obspy import UTCDateTime


class FeatureBaseClass(object):

    __feature_name__ = "undefined"

    def __init__(self):
        self.feature = None
        self.trace = None

    def save(self, h5FileName):
        """
        Save data to disk as netcdf4 files. Filenames will
        be 'YYYYMMDD_{sitename}.nc' where the date is taking
        from the xarray attributes 'starttime' and 'endtime'.

        :param h5FileName: Directory to save files to.
        :type h5FileName: str
        :param overwrite: Overwrite instead of merge if the
                          file already exists.
        :type overwrite: boolean
        """
        if self.feature is None:
            msg = "self.feature is None"
            raise TypeError(msg)

        xarray2hdf5(self.feature, h5FileName)

    def compute(self):
        msg = "compute is not implemented yet"
        raise NotImplementedError(msg)



import logging
import os

import numpy as np
from obspy import UTCDateTime
import pandas as pd
from tqdm import tqdm
import xarray as xr

from zizou.util import xarray2hdf5


logger = logging.getLogger(__name__)


class AnomalyDetectionBaseClass(object):
    
    def __init__(self, features, stacks=None, transforms=None):
        self.features = features
        self.stacks = stacks if stacks is not None else dict()
        self.transforms = transforms if transforms is not None else dict()
    
    def fit(self, starttime, endtime):
        """
        Train the model.
        """
        msg = "'training' is not implemented yet"
        raise NotImplementedError(msg)

    def infer(self, starttime, endtime):
        """
        Use the model to infer.
        """
        msg = "'inference' is not implemented yet"
        raise NotImplementedError(msg)

    def write_model_parameters(self):
        """
        Write principal components or 
        model weights to file
        """
        msg = "'write_model_parameters' is not implemented yet"
        raise NotImplementedError(msg)
    
    def read_model_parameters(self):
        """
        Read principal components or 
        model weights from file
        """
        msg = "'read_model_parameters' is not implemented yet"
        raise NotImplementedError(msg)

    def write_hyperparameters(self):
        """
        Store things like random seed,
        learning rate, number of principal
        components etc.
        """
        msg = "'write_hyperparameters' is not implemented yet"
        raise NotImplementedError(msg)

    def compute_anomaly_index(self):
        """
        A scalar value between 0 and 1 indicating
        the degree of anomaly (1=most anomaluous).
        """
        msg = "'compute_anomaly_index' is not implemented yet"
        raise NotImplementedError(msg)

    def get_features(self, data, featurefile=None):
        try:
            feats = xr.open_dataarray(featurefile)
        except (FileNotFoundError, ValueError):
            features = []
            for _f in self.features:
                logger.info(f"Reading feature {_f}.") 
                feat = data(_f, self.stacks.get(_f, None))
                if _f == 'sonogram':
                    for c in range(feat.shape[0]):
                        da = feat[c].reset_coords(names=['sonofrequency'], drop=True)
                        feat_name = 'sonogram_{:d}'.format(c)
                        features.append(da.rename(feat_name))
                else:
                    if self.transforms.get(_f, None) == 'log':
                        features.append(np.log10(feat))
                    else:
                        features.append(feat)
                
            feats = xr.merge(features).to_array()
            if featurefile is not None:
                feats.to_netcdf(featurefile)
        return feats.transpose('datetime', 'variable')

    def save(self, savedir: str):
        """
        Save data to disk as netcdf4 files. Filenames will
        be 'YYYYMMDD_{sitename}.nc' where the date is taking
        from the xarray attributes 'starttime' and 'endtime'.

        :param savedir: Directory to save files to.
        :type savedir: str
        """
        if self.classifications is None:
            msg = "No classifications found."
            raise TypeError(msg)

        xarray2hdf5(self.classifications, savedir)




import datetime
import inspect
from io import StringIO
import json
import os
import shutil
import tempfile
import unittest
from unittest.mock import patch, Mock

import pytest
import boto3
from moto import mock_s3
import numpy as np
import pandas as pd
from obspy import UTCDateTime, read_inventory, Trace
import xarray as xr
from zizou.util import xarray2hdf5, generate_test_data, test_signal


from zizou.data import (DataSource,
                       VolcanoMetadata,
                       FeatureRequest,
                       PostProcess,
                       PostProcessException)
from zizou import get_data


class DataTestCase(unittest.TestCase):

    def get_max(self, trace):
        value = np.nanmax(trace.data)
        _min = np.nanmin(trace.data)
        if abs(_min) > abs(value):
            value = _min
        return value

    def setUp(self):
        filename = inspect.getfile(inspect.currentframe())
        filedir = os.path.dirname(os.path.abspath(filename))
        self.sds_dir = os.path.join(filedir, "data", "sds_test")

    def test_exceptions(self):
        """
        Test that the right exceptions are raised.
        """
        with self.assertRaises(ValueError):
            ds = DataSource(sources=())

        with self.assertRaises(ValueError):
            ds = DataSource(sources=('blub', 'blab'))

        with self.assertRaises(ValueError):
            ds = DataSource(fdsn_urls=('/foo/bar'))

    @unittest.skip("Needs rewrite as it's testing a real FDSN server")
    def test_get_waveforms(self):
        """
        Test that the waveforms returned from SDS and FDSNws are
        consistent.
        """
        ds1 = DataSource(sources=('fdsn',))
        ds2 = DataSource(sources=('sds',), sds_dir=self.sds_dir,
                         staxml_dir=self.sds_dir)
        # dates are inclusive, and complete days
        starttime = UTCDateTime(year=2013, julday=240)
        endtime = starttime + 86400.
        for tr1 in ds1.get_waveforms('NZ', 'WIZ', '10', 'HHZ', starttime,
                                     endtime):
            for tr2 in ds2.get_waveforms('NZ', 'WIZ', '10', 'HHZ', starttime,
                                         endtime):
                self.assertEqual(tr1.max(), self.get_max(tr2))
                self.assertEqual(tr1.stats.npts, tr2.stats.npts)

    def test_with_gaps(self):
        """
        Test behaviour of traces with gaps.
        """
        ds1 = DataSource(sources=('sds',), sds_dir=self.sds_dir,
                         staxml_dir=self.sds_dir)
        starttime = UTCDateTime(2010, 11, 26, 0, 0, 0)
        endtime = starttime + 86400.
        stream = 'NZ.WIZ.10.HHZ'
        net, site, loc, comp = stream.split('.')
        gen = ds1.get_waveforms(net, site, loc, comp, starttime, endtime)
        t1 = next(gen)
        self.assertTrue(np.any(np.isnan(t1.data)))
        idx = np.where(np.isnan(t1.data))
        self.assertEqual(idx[0].size, 431238)

        starttime2 = UTCDateTime(2010, 11, 26, 0, 0, 0)
        endtime2 = starttime2 + 86400.
        gen2 = ds1.get_waveforms(net, site, loc, comp, starttime, endtime)
        t2 = next(gen2)

        # Check that we can take a trace apart and put it
        # together again
        t3 = t2.copy()
        mdata = np.ma.masked_invalid(t3.data)
        t3.data = mdata
        st = t3.split()
        st.merge(fill_value=np.nan)
        st.trim(starttime2, endtime2, pad=True, fill_value=np.nan,
                nearest_sample=False)
        idx = np.where(st[0].data != t2.data)
        # data 'differs' only at NaNs
        self.assertTrue(np.alltrue(np.isnan(st[0].data[idx])))

    @unittest.skip("Needs rewrite as it's testing a real FDSN server")
    def test_start_endtime(self):
        """
        Test that start and end times are what we expect.
        """
        ds = DataSource(sources=('sds',), sds_dir=self.sds_dir,
                        staxml_dir=self.sds_dir)
        starttime = UTCDateTime(year=2013, julday=240)
        endtime = starttime + 86400.
        stream = 'NZ.WIZ.10.HHZ'
        net, site, loc, comp = stream.split('.')
        gen = ds.get_waveforms(net, site, loc, comp, starttime, endtime)
        tr = next(gen)
        self.assertTrue(tr.stats.starttime >= starttime)
        self.assertTrue(tr.stats.endtime <= endtime)
        ds1 = DataSource(sources=('fdsn',))
        gen1 = ds1.get_waveforms(net, site, loc, comp, starttime, endtime)
        tr1 = next(gen1)
        self.assertTrue(tr1.stats.starttime >= starttime)
        self.assertTrue(tr1.stats.endtime <= endtime)
 
    @pytest.mark.slow
    def test_missing_station_metadata_sds(self):
        """
        At NTVZ, station metadata stops at
        2019-04-17T01:10:00.000000Z and restarts at
        2019-04-19T01:30:00.000000Z, but waveform data
        still exist throughout.

        Full get_waveform test across three days, which checks that traces
        are available for the time periods where there are both station
        and waveform data, np.nan otherwise.
        """
        ds = DataSource(sources=('sds',), sds_dir=self.sds_dir,
                        staxml_dir=self.sds_dir, chunk_size=86400.)
        starttime = UTCDateTime(2019, 4, 17)
        endtime = UTCDateTime(2019, 4, 20)
        net, site, loc, comp = ('NZ', 'NTVZ', '10', 'HHZ')
        gen = ds.get_waveforms(net, site, loc, comp, starttime, endtime)
        # number of non-NaN entries in each trace
        # 2019-04-17
        tr1 = next(gen)
        self.assertEqual(np.count_nonzero(~np.isnan(tr1.data)), 371902)
        # 2019-04-18
        tr2 = next(gen)
        self.assertEqual(np.count_nonzero(~np.isnan(tr2.data)), 0)
        # 2019-04-19
        tr3 = next(gen)
        self.assertEqual(np.count_nonzero(~np.isnan(tr3.data)), 8100000)
        
    @pytest.mark.slow
    @mock_s3
    def test_aws_backend(self):
        """
        Test retrieving waveform data from AWS S3.
        """
        ### Setup virtual S3
        bucket_name = 'virtual-geonet-open-data'
        client = boto3.client('s3', region_name='us-east-1',
                              aws_access_key_id='fake_access_key',
                              aws_secret_access_key='fake_secret_key')
        client.create_bucket(Bucket=bucket_name)
        starttime1 = UTCDateTime(2019, 4, 17, 11, 0, 17)
        endtime1 = UTCDateTime(2019, 4, 17, 23, 59, 59, 999999)
        starttime2 = UTCDateTime(2019, 4, 18, 0, 0, 0)
        endtime2 = UTCDateTime(2019, 4, 18, 3, 4, 5)
        key1 = "waveforms/miniseed/2019/2019.107/MAVZ.NZ/2019.107.MAVZ.10-HHZ.NZ.D"
        key2 = "waveforms/miniseed/2019/2019.108/MAVZ.NZ/2019.108.MAVZ.10-HHZ.NZ.D"
        test_data1 = test_signal(starttime=starttime1,
                                nsec=endtime1-starttime1,
                                station='MAVZ',
                                location='10') 
        test_data2 = test_signal(starttime=starttime2,
                                nsec=endtime2-starttime2,
                                station='MAVZ',
                                location='10') 
        filename = tempfile.mktemp()
        test_data1.write(filename, format='MSEED')
        client.upload_file(Filename=filename, Bucket=bucket_name, Key=key1)
        filename = tempfile.mktemp()
        test_data2.write(filename, format='MSEED')
        client.upload_file(Filename=filename, Bucket=bucket_name, Key=key2)
        
        ds = DataSource(staxml_dir=self.sds_dir, sources=('s3',),
                        s3bucket=bucket_name)
        net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
        gen = ds.get_waveforms(net, site, loc, comp, starttime1,
                               endtime2)
        tr1 = next(gen)
        test_data1.trim(starttime1, starttime1 + datetime.timedelta(days=1))
        test_inv = read_inventory(os.path.join(self.sds_dir, 'MAVZ.xml'))
        test_data1.attach_response(test_inv)
        test_data1.remove_sensitivity()
        self.assertEqual(test_data1.data.max(), np.nanmax(tr1.data))
        with self.assertRaises(StopIteration):
            tr2 = next(gen)
        self.assertEqual(tr1.stats.starttime, starttime1)
        self.assertEqual(tr1.stats.endtime, endtime2)
        gen1 = ds.get_waveforms(net, 'WIZ', loc, comp, starttime1,
                                endtime2)
        tr3 = next(gen1)
        # Check that the process returned a dummy trace
        self.assertEqual(len(tr3), 0)

    def test_mock_client(self):
        """
        Test that the mock client is working.
        """
        sds_dir = tempfile.mkdtemp()
        ds = DataSource(sds_dir=sds_dir, sources=['mock'])
        net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
        starttime = UTCDateTime(2024, 4, 11, 3, 0, 0)
        endtime = UTCDateTime(2024, 4, 11, 4, 0, 0)
        gen = ds.get_waveforms(net, site, loc, comp, starttime, endtime)
        tr = next(gen)
        self.assertEqual(tr.stats.starttime, starttime)
        self.assertEqual(tr.stats.endtime, endtime)
       

class PostProcessingTestCase(unittest.TestCase):
    """
    Tests various permutations of station metadata availability e.g. at
    MAVZ:
    [Channel 'HHZ', Location '10',
     Time range: 2012-05-22T02:00:00.000000Z - 2013-02-26T00:00:00.000000Z,
     Channel 'HHZ', Location '10',
     Time range: 2013-03-22T00:00:00.000000Z - 2013-04-09T22:15:00.000000Z,
     Channel 'HHZ', Location '10',
     Time range: 2013-04-10T00:15:00.000000Z - 2015-01-09T00:00:00.000000Z,
     Channel 'HHZ', Location '10',
     Time range: 2015-01-09T00:00:01.000000Z - 2017-04-20T22:00:00.000000Z,
     Channel 'HHZ', Location '10',
     Time range: 2017-04-20T22:00:02.000000Z - 9999-01-01T00:00:00.000000Z,
    ]
    """
    def setUp(self):
        filename = inspect.getfile(inspect.currentframe())
        filedir = os.path.dirname(os.path.abspath(filename))
        self.sds_dir = os.path.join(filedir, "data", "sds_test")
        self.ds = DataSource(sources=('sds',), sds_dir=self.sds_dir,
                             staxml_dir=self.sds_dir)

    def test_function_list(self):
        """
        Tests whether there is a matching number of tests and
        post-processing functions
        """
        pp = PostProcess()
        check_names = [check_n for (check_n, check) in pp.checklist]
        func_names = [func_n for (func_n, func) in pp.pp_functions]
        # Check correct number of checks and output function
        self.assertEqual(len(check_names), len(func_names))
        # Check checks and output functions match
        [self.assertEqual(c.replace('check', 'output'), f)
         for (c, f) in zip(check_names, func_names)]

    def test_case_single_trace_in_stream(self):
        net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
        startdate = UTCDateTime(2013, 2, 25)
        enddate = UTCDateTime(2013, 2, 26)
        st = self.ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                                     enddate, dtype='float64')
        inv = read_inventory(os.path.join(self.sds_dir, 'MAVZ.xml'),
                             format='STATIONXML')
        inv_dt = inv.select(location=loc, channel=comp,
                            starttime=startdate, endtime=enddate)
        pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                         comp=comp, fill_value=np.nan, startdate=startdate,
                         enddate=enddate)
        pp.run_checks()
        # Check correct test case:
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.checklist, pp.res) if b],
            ["check_case_no_station_issues"]
        )
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.pp_functions, pp.res) if b],
            ["output_case_no_station_issues"]
        )
        # Check output
        test_st = st.copy()
        test_st.attach_response(inv_dt)
        test_st.remove_sensitivity()
        test_st.trim(
            starttime=startdate, endtime=enddate,
            nearest_sample=False, pad=True, fill_value=np.nan
        )
        tr = pp.run_post_processing()
        np.testing.assert_array_almost_equal(tr.data, test_st[0].data)

    def test_multiple_traces_in_stream(self):
        net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
        startdate = UTCDateTime(2012, 5, 23)
        enddate = UTCDateTime(2012, 5, 24)
        st = self.ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                                     enddate, dtype='float64')
        inv = read_inventory(os.path.join(self.sds_dir, 'MAVZ.xml'),
                             format='STATIONXML')
        inv_dt = inv.select(location=loc, channel=comp,
                            starttime=startdate, endtime=enddate)
        pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                         comp=comp, fill_value=np.nan, startdate=startdate,
                         enddate=enddate)
        pp.run_checks()

        self.assertEqual(
            [a[0] for (a, b) in zip(pp.checklist, pp.res) if b],
            ["check_case_no_station_issues"]
        )
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.pp_functions, pp.res) if b],
            ["output_case_no_station_issues"]
        )
        tr = pp.run_post_processing()
        self.assertEqual(int(tr.stats.npts), int(8640000))

    def test_case_stream_starttime_equals_channel_endtime(self):
        net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
        startdate = UTCDateTime(2013, 2, 26)
        enddate = UTCDateTime(2013, 2, 27)
        st = self.ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                                     enddate, dtype='float64')
        inv = read_inventory(os.path.join(self.sds_dir, 'MAVZ.xml'),
                             format='STATIONXML')
        inv_dt = inv.select(location=loc, channel=comp,
                            starttime=startdate, endtime=enddate)
        pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                         comp=comp, fill_value=np.nan, startdate=startdate,
                         enddate=enddate)
        pp.run_checks()

        # Check correct test case:
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.checklist, pp.res) if b],
            ["check_case_no_data"]
        )
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.pp_functions, pp.res) if b],
            ["output_case_no_data"]
        )
        with self.assertRaises(PostProcessException):
            pp.run_post_processing()

    def test_case_stream_endtime_equals_channel_starttime(self):
        net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
        startdate = UTCDateTime(2013, 3, 21)
        enddate = UTCDateTime(2013, 3, 22)
        st = self.ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                                     enddate, dtype='float64')
        inv = read_inventory(os.path.join(self.sds_dir, 'MAVZ.xml'),
                             format='STATIONXML')
        inv_dt = inv.select(location=loc, channel=comp,
                            starttime=startdate, endtime=enddate)
        pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                         comp=comp, fill_value=np.nan, startdate=startdate,
                         enddate=enddate)
        pp.run_checks()

        # Check correct test case:
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.checklist, pp.res) if b],
            ["check_case_no_data"]
        )
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.pp_functions, pp.res) if b],
            ["output_case_no_data"]
        )
        with self.assertRaises(PostProcessException):
            pp.run_post_processing()

    def test_case_multiple_channel_periods(self):
        net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
        startdate = UTCDateTime(2015, 1, 9)
        enddate = UTCDateTime(2015, 1, 10)
        st = self.ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                                     enddate, dtype='float64')
        inv = read_inventory(os.path.join(self.sds_dir, 'MAVZ.xml'),
                             format='STATIONXML')
        inv_dt = inv.select(location=loc, channel=comp,
                            starttime=startdate, endtime=enddate)
        pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                         comp=comp, fill_value=np.nan, startdate=startdate,
                         enddate=enddate)
        pp.run_checks()

        # Check correct test case:
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.checklist, pp.res) if b],
            ["check_case_multiple_channel_periods"]
        )
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.pp_functions, pp.res) if b],
            ["output_case_multiple_channel_periods"]
        )

    def test_case_incomplete_station_metadata(self):
        startdate = UTCDateTime(2013, 4, 9)
        enddate = UTCDateTime(2013, 4, 10)
        net, site, loc, comp = ('NZ', 'MAVZ', '10', 'HHZ')
        inv = read_inventory(os.path.join(self.sds_dir, 'MAVZ.xml'),
                             format='STATIONXML')
        inv_dt = inv.select(location=loc, channel=comp,
                            starttime=startdate, endtime=enddate)
        st = self.ds.clients[0].client.get_waveforms(net, site, loc, comp, startdate,
                                                     enddate, dtype='float64')
        pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt, loc=loc,
                         comp=comp, fill_value=np.nan, startdate=startdate,
                         enddate=enddate)
        pp.run_checks()

        # Check correct test case:
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.checklist, pp.res) if b],
            ["check_case_incomplete_station_metadata"]
        )
        self.assertEqual(
            [a[0] for (a, b) in zip(pp.pp_functions, pp.res) if b],
            ["output_case_incomplete_station_metadata"]
        )
        test_st = st.copy()
        test_st.trim(startdate, inv_dt[0][0][0].end_date,
                     nearest_sample=False)
        test_st.attach_response(inv_dt)
        test_st.remove_sensitivity()
        test_st.trim(startdate, enddate, nearest_sample=False,
                     pad=True, fill_value=np.nan
                     )
        tr = pp.run_post_processing()

        np.testing.assert_array_almost_equal(tr.data, test_st[0].data)


class MetadataTestCase(unittest.TestCase):

    def setUp(self):
        self.data_dir = os.path.join(os.path.dirname(os.path.abspath(
            inspect.getfile(inspect.currentframe()))), "data")
        self.volcano = 'Ruapehu'
        self.eruptions = [
            datetime.date(2006, 10, 4),
            datetime.date(2007, 9, 25)
        ]
        self.st_names = [
            "NZ.MAVZ.10.HHZ",
            "NZ.WHVZ.10.HHZ",
            "NZ.TRVZ.10.HHZ",
            "NZ.FWVZ.10.HHZ",
            "NZ.COVZ.10.HHZ",
        ]

    def test_get_streams(self):
        """
        Test that the streams returned from VolcanoMetadata class are
        consistent.
        """
        file = get_data('package_data/config.yml')
        vm = VolcanoMetadata(file)
        streams = vm.get_available_streams(self.volcano)
        for stream in streams:
            self.assertTrue(stream in self.st_names)

    def test_get_eruptions(self):
        """
        Test that the eruptions dates returned from VolcanoMetadata class
        are consistent.
        """
        file = get_data('package_data/config.yml')
        vm = VolcanoMetadata(file)
        er = vm.get_eruption_dates(self.volcano)
        for e in er:
            self.assertTrue(e in self.eruptions)

    def test_find_station_coordinates(self):
        # Sample data for testing
        test_data = """ 
        metadata:
          volcano: 
          - name: Test Volcano
            network:
            - net_code: TZ
              stations:
              - sta_code: TST1
                latitude: -37.000
                longitude: 177.000
              - sta_code: TST2
                latitude: -38.000
                longitude: 178.000
        """
        vm = VolcanoMetadata(test_data)

        # Test case for an existing station
        assert vm.get_site_information("TST1")['latitude'] == -37.0
        assert vm.get_site_information("TST1")['longitude'] == 177.0

        # Test case for a non-existing station
        assert vm.get_site_information("NONEXISTENT") is None

class FeatureRequestTestCase(unittest.TestCase):

    def setUp(self):
        # create test directory
        self.tempdir = tempfile.mkdtemp()
        self.rootdir = os.path.join(self.tempdir, 'Mt_Doom', 'MDR', 'HHZ')
        try:
            os.makedirs(self.rootdir)
        except FileExistsError:
            pass

    def tearDown(self):
        if os.path.isdir(self.tempdir):
            shutil.rmtree(self.tempdir)
            
    def test_false_feature(self):
        fq = FeatureRequest(rootdir=self.tempdir)
        with self.assertRaises(ValueError):
            fakefeature = fq('fakefeature')

    def test_call_multiple_days(self):
        startdate = UTCDateTime(2016, 1, 1)._get_datetime()
        enddate = UTCDateTime(2016, 1, 2, 12)._get_datetime()
        xdf = generate_test_data(dim=1, ndays=20, tstart=startdate)
        xarray2hdf5(xdf, self.rootdir)
        fq = FeatureRequest(rootdir=self.tempdir)
        fq.starttime = startdate
        fq.endtime = enddate
        fq.volcano = 'Mt_Doom'
        fq.site = 'MDR'
        fq.channel = 'HHZ'
        rsam = fq('rsam')
        # Check data
        xd_index = dict(datetime=slice(startdate,
                                       (enddate-pd.to_timedelta(fq.interval))))

        np.testing.assert_array_almost_equal(
            rsam.loc[startdate:enddate], xdf.rsam.loc[xd_index], 5)
        # Check datetime range is correct
        first_time = pd.to_datetime(rsam.datetime.values[0])
        last_time = pd.to_datetime(rsam.datetime.values[-1])
        self.assertEqual(pd.to_datetime(startdate), first_time)
        self.assertEqual(pd.to_datetime(enddate) - pd.to_timedelta(fq.interval),
                         last_time)

    def test_call_single_day(self):
        fq = FeatureRequest(rootdir=self.rootdir)
        startdate = datetime.datetime(2016, 1, 2, 1)
        enddate = datetime.datetime(2016, 1, 2, 12)
        xdf = generate_test_data(dim=1, tstart=startdate)
        xarray2hdf5(xdf, self.rootdir)
        fq = FeatureRequest(rootdir=self.tempdir)
        fq.starttime = startdate
        fq.endtime = enddate
        fq.volcano = 'Mt_Doom'
        fq.site = 'MDR'
        fq.channel = 'HHZ'
        rsam = fq('rsam')
        # Check datetime range is correct
        first_time = pd.to_datetime(rsam.datetime.values[0])
        last_time = pd.to_datetime(rsam.datetime.values[-1])
        self.assertEqual(pd.to_datetime(startdate), first_time)
        self.assertEqual(pd.to_datetime(enddate) - pd.to_timedelta(fq.interval),
                         last_time)

    def test_rolling_window(self):
        startdate = UTCDateTime(2016, 1, 1)._get_datetime()
        enddate = UTCDateTime(2016, 1, 2, 12)._get_datetime()
        xdf = generate_test_data(dim=1, ndays=20, tstart=startdate)
        xarray2hdf5(xdf, self.rootdir)
        fq = FeatureRequest(rootdir=self.tempdir)

        stack_len_seconds = 3600
        stack_len_string = '1H'

        num_windows = int(stack_len_seconds / pd.Timedelta(fq.interval).seconds)
        fq.starttime = startdate
        fq.endtime = enddate
        fq.volcano = 'Mt_Doom'
        fq.site = 'MDR'
        fq.channel = 'HHZ'
        rsam = fq('rsam')
        rsam_rolling = fq('rsam', stack_length=stack_len_string)

        # Check correct datetime array
        np.testing.assert_array_equal(rsam.datetime.values,
                                      rsam_rolling.datetime.values)
        # Check correct values
        rolling_mean = [
            np.nanmean(rsam.data[(ind-num_windows+1):ind+1])
            for ind in np.arange(num_windows, len(rsam_rolling.data))
        ]
        np.testing.assert_array_almost_equal(
            np.array(rolling_mean), rsam_rolling.values[num_windows:], 6
        )

    def test_cache(self):
        """
        Test that data loaded from files and from cache are identical.
        """
        startdate = UTCDateTime(2016, 1, 1)._get_datetime()
        enddate = UTCDateTime(2016, 1, 2, 12)._get_datetime()
        xdf = generate_test_data(dim=1, ndays=20, tstart=startdate)
        xarray2hdf5(xdf, self.rootdir)
        fq = FeatureRequest(rootdir=self.tempdir)
        fq.starttime = startdate 
        fq.endtime = enddate 
        fq.volcano = 'Mt_Doom'
        fq.site = 'MDR'
        fq.channel = 'HHZ'
        data_from_files = fq('RSAM')
        data_from_cache = fq('RSAM')
        xr.testing.assert_equal(data_from_files, data_from_cache)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
Complex trace Analysis

"""
from configparser import ConfigParser

from scipy import signal
import scipy.integrate
from obspy.signal import util

import numpy as np
from obspy.core import UTCDateTime
import pandas as pd

from zizou import FeatureBaseClass


class Inst_Band(FeatureBaseClass):
    """
    Computes the instantaneous bandwidth of the given data which can be
    windowed or not. The instantaneous bandwidth is determined by the time
    derivative of the envelope normalized by the envelope of the input
    data.

    :param fil_coef: Filter coefficients for computing time derivative.
    :type fil_coef: :class:`~numpy.ndarray
    :return: **DataFrame** - Time,Instantaneous bandwidth of input data
    """

    def __init__(self, interval=600., fil_coef=(None, None, None),
                 configfile=None):
        super(Inst_Band, self).__init__()
        self.interval = interval
        self.fil_coef = fil_coef
        if configfile is not None:
            c = ConfigParser()
            c.read(configfile)
            self.interval = float(c['DEFAULT']['interval'])
            self.fil_coef = c['Insta_Band']['fil_coef']
            fc1, fc2, fc3 = c['Insta_Band']['fil_coef'].split(',')
            self.fil_coef = (float(fc1), float(fc2), float(fc3))

    def compute(self, trace):
        """
        :param trace: The seismic waveform data
        :type trace: :class:`obspy.Trace`
        """
        self.trace = trace
        self.data = self.trace.data
        self.fs = self.trace.stats.sampling_rate

        # initialise dataframe
        instan_band = pd.DataFrame()

        if len(self.trace) < 1:
            msg = "Trace is empty."
            raise ValueError(msg)

        ti = self.trace.stats.starttime
        index = 0
        # loop through data in interval blocks
        while ti < self.trace.stats.endtime:
            trint = self.trace.slice(ti, ti + self.interval)
            duration = trint.stats.npts * trint.stats.delta
            if duration >= 500:
                if duration < self.interval:
                    trint = self.trace.slice(
                        self.trace.stats.endtime - self.interval,
                        self.trace.stats.endtime
                    )

                # Calculate the envelope for interval blocks

                nfft = util.next_pow_2(trint.data.shape[-1])
                a_cpx = np.zeros((trint.data.shape), dtype=np.complex64)
                a_abs = np.zeros((trint.data.shape), dtype=np.float64)

                a_cpx = signal.hilbert(trint.data, nfft)
                a_abs = abs(signal.hilbert(trint.data, nfft))

                x = (a_cpx, a_abs)

                # Calculate Instantaneous bandwidth of input data
                row = x[1]

                # faster alternative to calculate A_win_add
                a_win_add = np.hstack(
                    ([row[0]] * (np.size(self.fil_coef) // 2), row,
                     [row[np.size(row) - 1]] * (np.size(self.fil_coef) // 2)))
                t = signal.lfilter(self.fil_coef, 1, a_win_add)
                # correct start and end values
                t = t[np.size(self.fil_coef) - 1:np.size(t)]
                sigma = abs((t * self.fs) / (x[1] * 2 * np.pi))

                # Create a DataFrame for output
                inst_b = pd.DataFrame(sigma)
                inst_b.drop(inst_b.tail(-trint.stats.npts).index, inplace=True)

                data1 = inst_b.mean()
                tstr = pd.to_datetime(
                    UTCDateTime.strftime(ti, '%Y-%m-%dT%H:%M:%S')
                )
                df = pd.DataFrame(data1.values, index=[tstr])
                instan_band = instan_band.append(df)
                index += 1
            ti += self.interval
        instan_band.columns = ['inst_band']

        self.feature = instan_band
        self.trace = trace
        return instan_band


class Inst_Freq(FeatureBaseClass):
    """
    Computes the instantaneous frequency of the given data which can be
    windowed or not. The instantaneous frequency is determined by the time
    derivative of the analytic signal of the input data.

    :param fil_coef: Filter coefficients for computing time derivative.
    :type fil_coef: :class:`~numpy.ndarray
    :return: **DataFrame** - Time, Instantaneous frequency of input data
    .
    """

    def __init__(self, interval=600., fil_coef=(None, None, None),
                 configfile=None):
        super(Inst_Freq, self).__init__()
        self.interval = interval
        self.fil_coef = fil_coef
        if configfile is not None:
            c = ConfigParser()
            c.read(configfile)
            self.interval = float(c['DEFAULT']['interval'])
            self.fil_coef = c['Insta_Freq']['fil_coef']
            fc1, fc2, fc3 = c['Insta_Freq']['fil_coef'].split(',')
            self.fil_coef = (float(fc1), float(fc2), float(fc3))

    def compute(self, trace):
        """
        :param trace: The seismic waveform data
        :type trace: :class:`obspy.Trace`
        """
        self.trace = trace
        self.data = self.trace.data
        self.fs = self.trace.stats.sampling_rate

        # initialise dataframe
        instan_freq = pd.DataFrame()

        if len(self.trace) < 1:
            msg = "Trace is empty."
            raise ValueError(msg)

        ti = self.trace.stats.starttime
        index = 0
        # loop through data in interval blocks
        while ti < self.trace.stats.endtime:
            trint = self.trace.slice(ti, ti + self.interval)
            duration = trint.stats.npts * trint.stats.delta
            if duration >= 500:
                if duration < self.interval:
                    trint = self.trace.slice(
                        self.trace.stats.endtime - self.interval,
                        self.trace.stats.endtime
                    )

                # Calculate the envelope for interval blocks

                nfft = util.next_pow_2(trint.data.shape[-1])
                a_cpx = np.zeros((trint.data.shape), dtype=np.complex64)
                a_abs = np.zeros((trint.data.shape), dtype=np.float64)

                a_cpx = signal.hilbert(trint.data, nfft)
                a_abs = abs(signal.hilbert(trint.data, nfft))

                x = (a_cpx, a_abs)

                # Calculate Instantaneous bandwidth of input data
                omega = np.zeros(np.size(x[0]), dtype=np.float64)
                f = np.real(x[0])
                h = np.imag(x[0])
                # faster alternative to calculate f_add
                f_add = np.hstack(
                    ([f[0]] * (np.size(self.fil_coef) // 2), f,
                     [f[np.size(f) - 1]] * (np.size(self.fil_coef) // 2)))
                fd = signal.lfilter(self.fil_coef, 1, f_add)
                # correct start and end values of time derivative
                fd = fd[np.size(self.fil_coef) - 1:np.size(fd)]
                # faster alternative to calculate h_add
                h_add = np.hstack(
                    ([h[0]] * (np.size(self.fil_coef) // 2), h,
                     [h[np.size(h) - 1]] * (np.size(self.fil_coef) // 2)))
                hd = signal.lfilter(self.fil_coef, 1, h_add)
                # correct start and end values of time derivative
                hd = hd[np.size(self.fil_coef) - 1:np.size(hd)]
                omega = abs(((f * hd - fd * h) / (f * f + h * h)) *
                            self.fs / 2 / np.pi)

                # Create a DataFrame for output

                inst_f = pd.DataFrame(omega)
                inst_f.drop(inst_f.tail(-trint.stats.npts).index, inplace=True)

                data1 = inst_f.mean()
                tstr = pd.to_datetime(
                    UTCDateTime.strftime(ti, '%Y-%m-%dT%H:%M:%S')
                )
                df = pd.DataFrame(data1.values, index=[tstr])
                instan_freq = instan_freq.append(df)
                index += 1
            ti += self.interval
        instan_freq.columns = ['inst_freq']

        self.feature = instan_freq
        self.trace = trace
        return instan_freq

  
class Norm_Env(FeatureBaseClass):
    """
    Computes the normalized envelope of the given data which can be
    windowed or not. In order to obtain a normalized measure of the signal
    envelope the instantaneous bandwidth of the smoothed envelope is
    normalized by the Nyquist frequency and is integrated afterwards.

    The time derivative of the normalized envelope is returned if input
    data are windowed only.

      :param fil_coef: Filter coefficients for computing time derivative.
      :type fil_coef: :class:`~numpy.ndarray
      :return: **DataFrame** - Time , Normalized envelope of input data
    """
    def __init__(self, window=10, interval=600., fil_coef=(None, None, None),
                 configfile=None):
        super(Norm_Env, self).__init__()
        self.interval = interval
        self.fil_coef = fil_coef
        self.window = window
        if configfile is not None:
            c = ConfigParser()
            c.read(configfile)
            self.interval = float(c['DEFAULT']['interval'])
            self.window = c['Norm_Env']['window']
            self.fil_coef = c['Norm_Env']['fil_coef']
            fc1, fc2, fc3 = c['Norm_Env']['fil_coef'].split(',')
            self.fil_coef = (float(fc1), float(fc2), float(fc3))

    def compute(self, trace):
        """
        :param trace: The seismic waveform data
        :type trace: :class:`obspy.Trace`
        """
        self.trace = trace
        self.data = self.trace.data
        self.fs = self.trace.stats.sampling_rate

        # initialise dataframe
        norm_env = pd.DataFrame()

        if len(self.trace) < 1:
            msg = "Trace is empty."
            raise ValueError(msg)

        ti = self.trace.stats.starttime
        index = 0
        # loop through data in interval blocks
        while ti < self.trace.stats.endtime:
            trint = self.trace.slice(ti, ti + self.interval)
            duration = trint.stats.npts * trint.stats.delta
            if duration >= 500:
                if duration < self.interval:
                    trint = self.trace.slice(
                        self.trace.stats.endtime - self.interval,
                        self.trace.stats.endtime
                    )

                # Calculate the envelope for interval blocks

                nfft = util.next_pow_2(trint.data.shape[-1])
                a_cpx = np.zeros((trint.data.shape), dtype=np.complex64)
                a_abs = np.zeros((trint.data.shape), dtype=np.float64)

                a_cpx = signal.hilbert(trint.data, nfft)
                a_abs = abs(signal.hilbert(trint.data, nfft))

                x = (a_cpx, a_abs)

                smoothie = int(self.window)

                # Calculate the normalize envelope
                a_win_smooth = util.smooth(x[1], smoothie)
                # Differentiation of original signal, dA/dt
                # Better, because faster, calculation of A_win_add
                a_win_add = np.hstack(
                    ([a_win_smooth[0]] * (np.size(self.fil_coef) // 2),
                     a_win_smooth, [a_win_smooth[np.size(a_win_smooth) - 1]] *
                     (np.size(self.fil_coef) // 2)))
                t = signal.lfilter(self.fil_coef, 1, a_win_add)
                # correct start and end values of time derivative
                t = t[np.size(self.fil_coef) - 1:np.size(t)]
                a_win_smooth[a_win_smooth < 1] = 1
                t_ = t / (2. * np.pi * (a_win_smooth) * (self.fs / 2.0))
                # Integral within window
                t_ = scipy.integrate.cumtrapz(t_, dx=(1.0 / self.fs))
                t_ = np.concatenate((t_[0:1], t_))
                anorm = ((np.exp(np.mean(t_))) - 1) * 100

                # Create a DataFrame for output

                data1 = {'norm_env': anorm}
                tstr = pd.to_datetime(
                    UTCDateTime.strftime(ti, '%Y-%m-%dT%H:%M:%S')
                )
                df = pd.DataFrame(data1, index=[tstr])
                norm_env = norm_env.append(df)
                index += 1
            ti += self.interval
        norm_env.columns = ['norm_env']

        self.feature = norm_env
        self.trace = trace
        return norm_env


class Centroid(FeatureBaseClass):
    """
    Computes the centroid time of the given data which can be windowed or
    not. The centroid time is determined as the time in the processed
    window where 50 per cent of the area below the envelope is reached.

    The time derivative of the centroid time is returned if input data are
    windowed only.

     :param fil_coef: Filter coefficients for computing time derivative.
     :type fil_coef: :class:`~numpy.ndarray

     :return: **DataFrame** - Time , Centroid time input data
    """
    def __init__(self, interval=600., configfile=None):
        super(Centroid, self).__init__()
        self.interval = interval
        if configfile is not None:
            c = ConfigParser()
            c.read(configfile)
            self.interval = float(c['DEFAULT']['interval'])

    def compute(self, trace):
        """
        :param trace: The seismic waveform data
        :type trace: :class:`obspy.Trace`
        """
        self.trace = trace
        self.data = self.trace.data
        self.fs = self.trace.stats.sampling_rate

        # initialise dataframe
        centroid1 = pd.DataFrame()

        if len(self.trace) < 1:
            msg = "Trace is empty."
            raise ValueError(msg)

        ti = self.trace.stats.starttime
        index = 0
        # loop through data in interval blocks
        while ti < self.trace.stats.endtime:
            trint = self.trace.slice(ti, ti + self.interval)
            duration = trint.stats.npts * trint.stats.delta
            if duration >= 500:
                if duration < self.interval:
                    trint = self.trace.slice(
                        self.trace.stats.endtime - self.interval,
                        self.trace.stats.endtime
                    )

                # Calculate the envelope
                nfft = util.next_pow_2(trint.data.shape[-1])
                a_cpx = signal.hilbert(trint.data, nfft)
                a_abs = abs(signal.hilbert(trint.data, nfft))

                x = (a_cpx, a_abs)

                # Calculate the centroid
                centro = np.zeros(1, dtype=np.float64)
                # Integral within window
                half = 0.5 * sum(x[1])
                # Estimate energy centroid
                for k in range(2, np.size(x[1])):
                    t = sum(x[1][0:k])
                    if (t >= half):
                        frac = (half - (t - sum(x[1][0:k - 1]))) / \
                               (t - (t - sum(x[1][0:k - 1])))
                        centro = \
                            (float(k) + float(frac)) / float(np.size(x[1]))

                # Create a DataFrame for output1
                data1 = {'centroid': centro}
                tstr = pd.to_datetime(
                    UTCDateTime.strftime(ti, '%Y-%m-%dT%H:%M:%S')
                )
                df = pd.DataFrame(data1, index=[tstr])
                centroid1 = centroid1.append(df)
                index += 1
            ti += self.interval
        centroid1.columns = ['centroid']
        self.feature = centroid1
        self.trace = trace
        return centroid1


class WindowCentroid(FeatureBaseClass):
    """
    Computes the centroid of a given time window,
    defined between 0 and 1.
    Deltas at t_start, t_end and t_midpoint give
    values of 0, 1 and 0.5 respectively.
    A flat signal has a centroid = 0.57.

    :param interval: Length of the time interval
                     to compute feature.
    :type interval: int, optional
    """

    def __init__(self, interval=600., configfile=None):
        super(WindowCentroid, self).__init__()
        self.interval = interval
        if configfile is not None:
            c = ConfigParser()
            c.read(configfile)
            self.interval = float(c['DEFAULT']['interval'])

    def compute(self, trace):
        if len(trace) < 1:
            msg = "Trace is empty."
            raise ValueError(msg)
        tr = trace.copy()
        starttime = tr.stats.starttime
        endtime = tr.stats.endtime
        # Filter settings
        f1, f2 = 0.1, 20
        # handle gaps (i.e. NaNs)
        mdata = np.ma.masked_invalid(tr.data)
        tr.data = mdata
        st = tr.split()
        # Filter
        st.detrend('linear')
        st.detrend('constant')
        st.filter('bandpass', freqmin=f1, freqmax=f2, corners=4,
                  zerophase=False)
        st.merge(fill_value=np.nan)
        st.trim(starttime, endtime, pad=True, fill_value=np.nan)
        tr = st[0]
        tr.taper(0.025)
        tr.differentiate()

        # initialise dataframe
        window_centroid = pd.DataFrame()

        t = tr.stats.starttime
        index = 0
        # loop through data in interval blocks
        while t < tr.stats.endtime:
            trint = tr.slice(t, t + self.interval)
            duration = trint.stats.npts * trint.stats.delta
            if duration >= 0.8 * self.interval:
                if duration < self.interval:
                    trint = tr.slice(trace.stats.endtime - self.interval,
                                     trace.stats.endtime)

                m0 = 2 * scipy.integrate.simps(trint.times() ** 0 * np.abs(trint.data))
                m2 = 2 * scipy.integrate.simps(trint.times() ** 2 * np.abs(trint.data))
                centroid = np.sqrt(m2 / m0) / trint.times()[-1]

                data = {'window_centroid': float(format(centroid, '.4f'))}
                ts = UTCDateTime.strftime(t, '%Y-%m-%dT%H:%M:%S')
                tstr = pd.to_datetime(ts)
                window_centroid = \
                    window_centroid.append(
                        pd.DataFrame(data, index=[tstr])
                    )
                index += 1
            t += self.interval
        self.feature = window_centroid
        self.trace = tr
        return window_centroid
